# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""General utilities for command line examples."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import namedtuple
from yaml import safe_load

from uber_rides.client import UberRidesClient
from uber_rides.session import OAuth2Credential
from uber_rides.session import Session


# set your app credentials here
CREDENTIALS_FILENAME = 'example/config.yaml'

# where your OAuth 2.0 credentials are stored
STORAGE_FILENAME = 'example/oauth2_session_store.yaml'

DEFAULT_CONFIG_VALUES = frozenset([
    'INSERT_CLIENT_ID_HERE',
    'INSERT_CLIENT_SECRET_HERE',
    'INSERT_REDIRECT_URL_HERE',
])

Colors = namedtuple('Colors', 'response, success, fail, end')
COLORS = Colors(
    response='\033[94m',
    success='\033[92m',
    fail='\033[91m',
    end='\033[0m',
)


def success_print(message):
    """Print a message in green text.

    Parameters
        message (str)
            Message to print.
    """
    print(COLORS.success, message, COLORS.end)


def response_print(message):
    """Print a message in blue text.

    Parameters
        message (str)
            Message to print.
    """
    print(COLORS.response, message, COLORS.end)


def fail_print(error):
    """Print an error in red text.

    Parameters
        error (HTTPError)
            Error object to print.
    """
    print(COLORS.fail, error.message, COLORS.end)
    print(COLORS.fail, error.errors, COLORS.end)


def paragraph_print(message):
    """Print message with padded newlines.

    Parameters
        message (str)
            Message to print.
    """
    paragraph = '\n{}\n'
    print(paragraph.format(message))


def import_app_credentials(filename=CREDENTIALS_FILENAME):
    """Import app credentials from configuration file.

    Parameters
        filename (str)
            Name of configuration file.

    Returns
        credentials (dict)
            All your app credentials and information
            imported from the configuration file.
    """
    with open(filename, 'r') as config_file:
        config = safe_load(config_file)

    client_id = config['client_id']
    client_secret = config['client_secret']
    redirect_url = config['redirect_url']

    config_values = [client_id, client_secret, redirect_url]

    for value in config_values:
        if value in DEFAULT_CONFIG_VALUES:
            exit('Missing credentials in {}'.format(filename))

    credentials = {
        'client_id': client_id,
        'client_secret': client_secret,
        'redirect_url': redirect_url,
        'scopes': set(config['scopes']),
    }

    return credentials


def import_oauth2_credentials(filename=STORAGE_FILENAME):
    """Import OAuth 2.0 session credentials from storage file.

    Parameters
        filename (str)
            Name of storage file.

    Returns
        credentials (dict)
            All your app credentials and information
            imported from the configuration file.
    """
    with open(filename, 'r') as storage_file:
        storage = safe_load(storage_file)

    # depending on OAuth 2.0 grant_type, these values may not exist
    client_secret = storage.get('client_secret')
    redirect_url = storage.get('redirect_url')
    refresh_token = storage.get('refresh_token')

    credentials = {
        'access_token': storage['access_token'],
        'client_id': storage['client_id'],
        'client_secret': client_secret,
        'expires_in_seconds': storage['expires_in_seconds'],
        'grant_type': storage['grant_type'],
        'redirect_url': redirect_url,
        'refresh_token': refresh_token,
        'scopes': storage['scopes'],
    }

    return credentials


def create_uber_client(credentials):
    """Create an UberRidesClient from OAuth 2.0 credentials.

    Parameters
        credentials (dict)
            Dictionary of OAuth 2.0 credentials.

    Returns
        (UberRidesClient)
            An authorized UberRidesClient to access API resources.
    """
    oauth2credential = OAuth2Credential(
        client_id=credentials.get('client_id'),
        access_token=credentials.get('access_token'),
        expires_in_seconds=credentials.get('expires_in_seconds'),
        scopes=credentials.get('scopes'),
        grant_type=credentials.get('grant_type'),
        redirect_url=credentials.get('redirect_url'),
        client_secret=credentials.get('client_secret'),
        refresh_token=credentials.get('refresh_token'),
    )
    session = Session(oauth2credential=oauth2credential)
    return UberRidesClient(session, sandbox_mode=True)



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from mock import Mock
from pytest import fixture

from uber_rides.session import OAuth2Credential
from uber_rides.session import Session
from uber_rides.utils import auth
from uber_rides.utils import http


# replace these with valid tokens and credentials to rerecord fixtures
CLIENT_ID = 'xxx'
CLIENT_SECRET = 'xxx'
SERVER_TOKEN = 'xxx'

ACCESS_TOKEN = 'xxx'
REFRESH_TOKEN = 'xxx'
REDIRECT_URL = 'https://uberapitester.com/api/v1/uber/oauth'

EXPIRES_IN_SECONDS = 3000

SCOPES_STRING = 'profile history'
SCOPES_SET = {'profile', 'history'}

CLIENT_CREDENTIALS_SCOPES_STRING = 'partner.referrals'
CLIENT_CREDENTIALS_SCOPES_SET = {'partner.referrals'}


@fixture
def server_token_session():
    """Create a Session with Server Token."""
    return Session(
        server_token=SERVER_TOKEN,
    )


@fixture
def authorization_code_grant_session():
    """Create a Session from Auth Code Grant Credential."""
    oauth2credential = OAuth2Credential(
        client_id=None,
        redirect_url=None,
        access_token=ACCESS_TOKEN,
        expires_in_seconds=EXPIRES_IN_SECONDS,
        scopes=SCOPES_SET,
        grant_type=auth.AUTHORIZATION_CODE_GRANT,
        client_secret=None,
        refresh_token=REFRESH_TOKEN,
    )

    return Session(oauth2credential=oauth2credential)


@fixture
def implicit_grant_session():
    """Create a Session from Implicit Grant Credential."""
    oauth2credential = OAuth2Credential(
        client_id=None,
        redirect_url=None,
        access_token=ACCESS_TOKEN,
        expires_in_seconds=EXPIRES_IN_SECONDS,
        scopes=SCOPES_SET,
        grant_type=auth.IMPLICIT_GRANT,
        client_secret=None,
        refresh_token=None,
    )

    return Session(oauth2credential=oauth2credential)


@fixture
def client_credential_grant_session():
    """Create a Session from Client Credential Grant."""
    oauth2credential = OAuth2Credential(
        client_id=None,
        redirect_url=None,
        access_token=ACCESS_TOKEN,
        expires_in_seconds=EXPIRES_IN_SECONDS,
        scopes=SCOPES_SET,
        grant_type=auth.CLIENT_CREDENTIAL_GRANT,
        client_secret=None,
        refresh_token=None,
    )

    return Session(oauth2credential=oauth2credential)


@fixture
def authorization_code_response():
    """Response after Authorization Code Access Request."""

    mock_response = Mock(
        status_code=http.STATUS_OK,
    )

    response_json = {
        'access_token': ACCESS_TOKEN,
        'expires_in': EXPIRES_IN_SECONDS,
        'scope': SCOPES_STRING,
        'refresh_token': REFRESH_TOKEN,
    }

    mock_response.json = Mock(return_value=response_json)
    return mock_response


@fixture
def client_credentials_response():
    """Response after Client Credentials Access Request."""
    mock_response = Mock(
        status_code=http.STATUS_OK,
    )

    response_json = {
        'access_token': ACCESS_TOKEN,
        'expires_in': EXPIRES_IN_SECONDS,
        'scope': CLIENT_CREDENTIALS_SCOPES_STRING,
    }

    mock_response.json = Mock(return_value=response_json)
    return mock_response


def test_server_token_session_initialized(server_token_session):
    """Confirm Session with Server Token initialized correctly."""
    assert server_token_session.server_token == SERVER_TOKEN
    assert server_token_session.token_type == auth.SERVER_TOKEN_TYPE
    assert server_token_session.oauth2credential is None


def test_oauth2_session_initialized(authorization_code_grant_session):
    """Confirm Session with OAuth2Credential initialized correctly."""
    assert authorization_code_grant_session.server_token is None
    assert authorization_code_grant_session.token_type == auth.OAUTH_TOKEN_TYPE
    oauth2 = authorization_code_grant_session.oauth2credential
    assert oauth2.access_token == ACCESS_TOKEN
    assert oauth2.scopes == SCOPES_SET
    assert oauth2.grant_type == auth.AUTHORIZATION_CODE_GRANT
    assert oauth2.refresh_token is REFRESH_TOKEN


def test_new_authorization_code_grant_session_is_not_stale(
    authorization_code_grant_session,
):
    """Confirm that a new Session from Auth Code Grant is not stale."""
    assert authorization_code_grant_session.server_token is None
    assert authorization_code_grant_session.oauth2credential
    assert authorization_code_grant_session.token_type == auth.OAUTH_TOKEN_TYPE
    assert not authorization_code_grant_session.oauth2credential.is_stale()


def test_new_implicit_grant_session_is_not_stale(implicit_grant_session):
    """Confirm that a new Session from Implicit Grant is not stale."""
    assert implicit_grant_session.server_token is None
    assert implicit_grant_session.oauth2credential
    assert implicit_grant_session.token_type == auth.OAUTH_TOKEN_TYPE
    assert not implicit_grant_session.oauth2credential.is_stale()


def test_new_client_credential_session_is_not_stale(
    client_credential_grant_session,
):
    """Confirm that a new Session from Client Credential Grant is not stale."""
    assert client_credential_grant_session.server_token is None
    assert client_credential_grant_session.oauth2credential
    assert client_credential_grant_session.token_type == auth.OAUTH_TOKEN_TYPE
    assert not client_credential_grant_session.oauth2credential.is_stale()


def test_old_authorization_code_grant_session_is_stale(
    authorization_code_grant_session
):
    """Confirm that an old Session from Auth Code Grant is stale."""
    authorization_code_grant_session.oauth2credential.expires_in_seconds = 1
    assert authorization_code_grant_session.oauth2credential.is_stale()


def test_old_implicit_grant_session_is_stale(implicit_grant_session):
    """Confirm that an old Session from Implicit Grant is stale."""
    implicit_grant_session.oauth2credential.expires_in_seconds = 1
    assert implicit_grant_session.oauth2credential.is_stale()


def test_old_client_credential_session_is_stale(
    client_credential_grant_session,
):
    """Confirm that an old Session from Client Credential Grant is stale."""
    client_credential_grant_session.oauth2credential.expires_in_seconds = 1
    assert client_credential_grant_session.oauth2credential.is_stale()


def test_make_session_from_authorization_code_response(
    authorization_code_response,
):
    """Test classmethod to build OAuth2Credential from HTTP Response."""
    oauth2credential = OAuth2Credential.make_from_response(
        response=authorization_code_response,
        grant_type=auth.AUTHORIZATION_CODE_GRANT,
        client_id=CLIENT_ID,
        client_secret=CLIENT_SECRET,
        redirect_url=REDIRECT_URL,
    )

    assert oauth2credential.access_token == ACCESS_TOKEN
    assert oauth2credential.scopes == SCOPES_SET
    assert oauth2credential.grant_type == auth.AUTHORIZATION_CODE_GRANT
    assert oauth2credential.refresh_token == REFRESH_TOKEN
    assert oauth2credential.client_id == CLIENT_ID
    assert oauth2credential.client_secret == CLIENT_SECRET
    assert oauth2credential.redirect_url == REDIRECT_URL


def test_make_session_from_client_credentials_response(
    client_credentials_response,
):
    """Test classmethod to build OAuth2Credential from HTTP Response."""
    oauth2credential = OAuth2Credential.make_from_response(
        response=client_credentials_response,
        grant_type=auth.CLIENT_CREDENTIAL_GRANT,
        client_id=CLIENT_ID,
        client_secret=CLIENT_SECRET,
    )

    assert oauth2credential.access_token == ACCESS_TOKEN
    assert oauth2credential.scopes == CLIENT_CREDENTIALS_SCOPES_SET
    assert oauth2credential.grant_type == auth.CLIENT_CREDENTIAL_GRANT
    assert oauth2credential.refresh_token is None
    assert oauth2credential.client_id == CLIENT_ID
    assert oauth2credential.client_secret == CLIENT_SECRET
    assert oauth2credential.redirect_url is None



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from mock import Mock
from pytest import fixture

from uber_rides.errors import ClientError
from uber_rides.errors import ErrorDetails
from uber_rides.errors import ServerError
from uber_rides.utils import http


@fixture
def simple_401_error():
    code = 'unauthorized'

    mock_error = Mock(
        status_code=http.STATUS_UNAUTHORIZED,
        headers=http.DEFAULT_CONTENT_HEADERS,
    )

    error_response = {
        'message': http.ERROR_CODE_DESCRIPTION_DICT[code],
        'code': code,
    }

    mock_error.json = Mock(return_value=error_response)
    return mock_error


@fixture
def simple_422_validation_error():
    code = 'validation_failed'

    mock_error = Mock(
        status_code=http.STATUS_UNPROCESSABLE_ENTITY,
        headers=http.DEFAULT_CONTENT_HEADERS,
    )

    error_response = {
        'fields': {
            'latitude': 'Must be between -90.0 and 90.0',
            'longitude': 'Must be between -180.0 and 180.0',
        },
        'message': http.ERROR_CODE_DESCRIPTION_DICT[code],
        'code': code,
    }

    mock_error.json = Mock(return_value=error_response)
    return mock_error


@fixture
def simple_422_distance_exceeded_error():
    code = 'distance_exceeded'

    mock_error = Mock(
        status_code=http.STATUS_UNPROCESSABLE_ENTITY,
        headers=http.DEFAULT_CONTENT_HEADERS,
    )

    error_response = {
        'fields': {
            'start_longitude': http.ERROR_CODE_DESCRIPTION_DICT[code],
            'end_longitude': http.ERROR_CODE_DESCRIPTION_DICT[code],
            'start_latitude': http.ERROR_CODE_DESCRIPTION_DICT[code],
            'end_latitude': http.ERROR_CODE_DESCRIPTION_DICT[code],
        },
        'message': http.ERROR_CODE_DESCRIPTION_DICT[code],
        'code': code,
    }

    mock_error.json = Mock(return_value=error_response)
    return mock_error


@fixture
def simple_500_error():
    code = 'internal_server_error'

    mock_error = Mock(
        status_code=http.STATUS_INTERNAL_SERVER_ERROR,
        headers=http.DEFAULT_CONTENT_HEADERS,
    )

    error_response = {
        'message': http.ERROR_CODE_DESCRIPTION_DICT[code],
        'code': code,
    }

    mock_error.json = Mock(return_value=error_response)
    return mock_error


@fixture
def simple_503_error():
    code = 'service_unavailable'

    mock_error = Mock(
        status_code=http.STATUS_SERVICE_UNAVAILABLE,
        headers=http.DEFAULT_CONTENT_HEADERS,
    )

    error_response = {
        'message': http.ERROR_CODE_DESCRIPTION_DICT[code],
        'code': code,
    }

    mock_error.json = Mock(return_value=error_response)
    return mock_error


@fixture
def complex_409_surge_error():
    code = 'surge'

    mock_error = Mock(
        status_code=http.STATUS_CONFLICT,
        headers=http.DEFAULT_CONTENT_HEADERS,
    )

    error_response = {
        'meta': {
            'surge_confirmation': {
                'href': 'api.uber.com/v1.2/surge-confirmations/abc',
                'surge_confirmation_id': 'abc',
            },
        },
        'errors': [
            {
                'status': http.STATUS_CONFLICT,
                'code': code,
                'title': http.ERROR_CODE_DESCRIPTION_DICT[code],
            },
        ],
    }

    mock_error.json = Mock(return_value=error_response)
    return mock_error


@fixture
def complex_422_same_pickup_dropoff_error():
    code = 'same_pickup_dropoff'

    mock_error = Mock(
        status_code=http.STATUS_UNPROCESSABLE_ENTITY,
        headers=http.DEFAULT_CONTENT_HEADERS,
    )

    error_response = {
        'meta': {},
        'errors': [
            {
                'status': http.STATUS_UNPROCESSABLE_ENTITY,
                'code': code,
                'title': http.ERROR_CODE_DESCRIPTION_DICT[code],
            },
        ],
    }

    mock_error.json = Mock(return_value=error_response)
    return mock_error


def test_simple_401_error(simple_401_error):
    """Test Unauthorized Error converted to ClientError correctly."""
    client_error = ClientError(simple_401_error, 'msg')

    assert str(client_error) == 'msg'
    assert isinstance(client_error.errors, list)
    assert isinstance(client_error.meta, dict)
    assert not client_error.meta

    error_details = client_error.errors[0]
    expected_code = 'unauthorized'
    expected_description = http.ERROR_CODE_DESCRIPTION_DICT[expected_code]

    assert isinstance(error_details, ErrorDetails)
    assert error_details.status == http.STATUS_UNAUTHORIZED
    assert error_details.code == expected_code
    assert error_details.title == expected_description


def test_simple_422_validation_error(simple_422_validation_error):
    """Test Validation Error converted to ClientError correctly."""
    client_error = ClientError(simple_422_validation_error, 'msg')

    assert str(client_error) == 'msg'
    assert isinstance(client_error.errors, list)
    assert isinstance(client_error.meta, dict)

    error_details = client_error.errors[0]
    expected_code = 'validation_failed'
    expected_description = http.ERROR_CODE_DESCRIPTION_DICT[expected_code]

    assert isinstance(error_details, ErrorDetails)
    assert error_details.status == http.STATUS_UNPROCESSABLE_ENTITY
    assert error_details.code == expected_code
    assert error_details.title == expected_description

    meta = client_error.meta

    fields = {
        'fields': {
            'latitude': 'Must be between -90.0 and 90.0',
            'longitude': 'Must be between -180.0 and 180.0',
        },
    }

    assert meta == fields


def test_simple_422_distance_exceeded_error(
    simple_422_distance_exceeded_error
):
    """Test Distance Exceeded Error converted to ClientError correctly."""
    client_error = ClientError(simple_422_distance_exceeded_error, 'msg')

    assert str(client_error) == 'msg'
    assert isinstance(client_error.errors, list)
    assert isinstance(client_error.meta, dict)

    error_details = client_error.errors[0]
    expected_code = 'distance_exceeded'
    expected_description = http.ERROR_CODE_DESCRIPTION_DICT[expected_code]

    assert isinstance(error_details, ErrorDetails)
    assert error_details.status == http.STATUS_UNPROCESSABLE_ENTITY
    assert error_details.code == expected_code
    assert error_details.title == expected_description

    meta = client_error.meta

    fields = {
        'fields': {
            'start_longitude': http.ERROR_CODE_DESCRIPTION_DICT[expected_code],
            'end_longitude': http.ERROR_CODE_DESCRIPTION_DICT[expected_code],
            'start_latitude': http.ERROR_CODE_DESCRIPTION_DICT[expected_code],
            'end_latitude': http.ERROR_CODE_DESCRIPTION_DICT[expected_code],
        },
    }

    assert meta == fields


def test_simple_500_error(simple_500_error):
    """Test Internal Server Error converted to ClientError correctly."""
    server_error = ServerError(simple_500_error, 'msg')

    assert str(server_error) == 'msg'
    assert isinstance(server_error.meta, dict)
    assert not server_error.meta

    error_details = server_error.error   # single error instead of array
    expected_code = 'internal_server_error'
    expected_description = http.ERROR_CODE_DESCRIPTION_DICT[expected_code]

    assert isinstance(error_details, ErrorDetails)
    assert error_details.status == http.STATUS_INTERNAL_SERVER_ERROR
    assert error_details.code == expected_code
    assert error_details.title == expected_description


def test_simple_503_error(simple_503_error):
    """Test Service Unavailable Error converted to ClientError correctly."""
    server_error = ServerError(simple_503_error, 'msg')

    assert str(server_error) == 'msg'
    assert isinstance(server_error.meta, dict)
    assert not server_error.meta

    error_details = server_error.error   # single error instead of array
    expected_code = 'service_unavailable'
    expected_description = http.ERROR_CODE_DESCRIPTION_DICT[expected_code]

    assert isinstance(error_details, ErrorDetails)
    assert error_details.status == http.STATUS_SERVICE_UNAVAILABLE
    assert error_details.code == expected_code
    assert error_details.title == expected_description


def test_complex_409_surge_error(complex_409_surge_error):
    """Test Surge Error converted to ClientError correctly."""
    client_error = ClientError(complex_409_surge_error, 'msg')

    assert str(client_error) == 'msg'
    assert isinstance(client_error.errors, list)
    assert isinstance(client_error.meta, dict)

    error_details = client_error.errors[0]
    expected_code = 'surge'
    expected_description = http.ERROR_CODE_DESCRIPTION_DICT[expected_code]

    assert isinstance(error_details, ErrorDetails)
    assert error_details.status == http.STATUS_CONFLICT
    assert error_details.code == expected_code
    assert error_details.title == expected_description

    surge_meta = client_error.meta['surge_confirmation']

    assert surge_meta['surge_confirmation_id'] == 'abc'
    assert surge_meta['href'] == 'api.uber.com/v1.2/surge-confirmations/abc'


def test_complex_422_same_pickup_dropoff_error(
    complex_422_same_pickup_dropoff_error
):
    """Test Same Pickup-Dropoff Error converted to ClientError correctly."""
    client_error = ClientError(complex_422_same_pickup_dropoff_error, 'msg')

    assert str(client_error) == 'msg'
    assert isinstance(client_error.errors, list)
    assert isinstance(client_error.meta, dict)
    assert not client_error.meta

    error_details = client_error.errors[0]
    expected_code = 'same_pickup_dropoff'
    expected_description = http.ERROR_CODE_DESCRIPTION_DICT[expected_code]

    assert isinstance(error_details, ErrorDetails)
    assert error_details.status == http.STATUS_UNPROCESSABLE_ENTITY
    assert error_details.code == expected_code
    assert error_details.title == expected_description


def test_error_details_dump():
    """Test the object representation repr method for ErrorDetails."""
    client_error = ErrorDetails(status='status', code=400, title='msg')

    assert str(client_error) == "ErrorDetails: {} {} {}".format(
        str(client_error.status),
        str(client_error.code),
        str(client_error.title)
    )



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Configure vcr to record test fixtures.

This file overrides vcrpy's default settings.
You can set options individually. Such per-cassette
overrides will take precedent over the global
configuration below.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import vcr

uber_vcr = vcr.VCR(
    serializer='yaml',
    cassette_library_dir='tests/fixtures',

    # disable gzip
    decode_compressed_response=True,

    # you can record_mode='all' to force re-record all cassettes
    record_mode='once',

    # configure matching features that vcr uses to identify identical requests
    match_on=['uri', 'method'],

    # scrub sensitive information from cassette files
    filter_headers=['Authorization'],
)



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Initializes an UberRidesClient with OAuth 2.0 Credentials.

This example demonstrates how to get an access token through the
OAuth 2.0 Authorization Code Grant and use credentials to create
an UberRidesClient.

To run this example:

    (1) Set your app credentials in config.yaml
    (2) Run `python authorization_code_grant.py`
    (3) A success message will print, 'Hello {YOUR_NAME}'
    (4) User OAuth 2.0 credentials are recorded in
        'oauth2_session_store.yaml'
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import input

from yaml import safe_dump

from example import utils
from example.utils import fail_print
from example.utils import response_print
from example.utils import success_print
from example.utils import import_app_credentials

from uber_rides.auth import AuthorizationCodeGrant
from uber_rides.client import UberRidesClient
from uber_rides.errors import ClientError
from uber_rides.errors import ServerError
from uber_rides.errors import UberIllegalState


def authorization_code_grant_flow(credentials, storage_filename):
    """Get an access token through Authorization Code Grant.

    Parameters
        credentials (dict)
            All your app credentials and information
            imported from the configuration file.
        storage_filename (str)
            Filename to store OAuth 2.0 Credentials.

    Returns
        (UberRidesClient)
            An UberRidesClient with OAuth 2.0 Credentials.
    """
    auth_flow = AuthorizationCodeGrant(
        credentials.get('client_id'),
        credentials.get('scopes'),
        credentials.get('client_secret'),
        credentials.get('redirect_url'),
    )

    auth_url = auth_flow.get_authorization_url()
    login_message = 'Login and grant access by going to:\n{}\n'
    login_message = login_message.format(auth_url)
    response_print(login_message)

    redirect_url = 'Copy the URL you are redirected to and paste here: \n'
    result = input(redirect_url).strip()

    try:
        session = auth_flow.get_session(result)

    except (ClientError, UberIllegalState) as error:
        fail_print(error)
        return

    credential = session.oauth2credential

    credential_data = {
        'client_id': credential.client_id,
        'redirect_url': credential.redirect_url,
        'access_token': credential.access_token,
        'expires_in_seconds': credential.expires_in_seconds,
        'scopes': list(credential.scopes),
        'grant_type': credential.grant_type,
        'client_secret': credential.client_secret,
        'refresh_token': credential.refresh_token,
    }

    with open(storage_filename, 'w') as yaml_file:
        yaml_file.write(safe_dump(credential_data, default_flow_style=False))

    return UberRidesClient(session, sandbox_mode=True)


def hello_user(api_client):
    """Use an authorized client to fetch and print profile information.

    Parameters
        api_client (UberRidesClient)
            An UberRidesClient with OAuth 2.0 credentials.
    """

    try:
        response = api_client.get_user_profile()

    except (ClientError, ServerError) as error:
        fail_print(error)
        return

    else:
        profile = response.json
        first_name = profile.get('first_name')
        last_name = profile.get('last_name')
        email = profile.get('email')
        message = 'Hello, {} {}. Successfully granted access token to {}.'
        message = message.format(first_name, last_name, email)
        success_print(message)
        success_print(profile)

        success_print('---')
        response = api_client.get_home_address()
        address = response.json
        success_print(address)

        success_print('---')
        response = api_client.get_user_activity()
        history = response.json
        success_print(history)


if __name__ == '__main__':
    """Run the example.

    Get an access token through the OAuth 2.0 Authorization Code Grant
    and use credentials to create an UberRidesClient.
    """
    credentials = import_app_credentials()

    api_client = authorization_code_grant_flow(
        credentials,
        utils.STORAGE_FILENAME,
    )

    hello_user(api_client)



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from requests import codes
from time import time

from uber_rides.errors import ClientError
from uber_rides.errors import UberIllegalState
from uber_rides.utils import auth


EXPIRES_THRESHOLD_SECONDS = 500


class Session(object):
    """A class to store credentials.

    A Session can be initialized with a Server Token or with a set of
    OAuth 2.0 Credentials, but not with both. A Session uses credentials
    to properly construct requests to Uber and access protected resources.
    """

    def __init__(
        self,
        server_token=None,
        oauth2credential=None,
    ):
        """Initialize a Session.

        Parameters
            sever_token (str)
                Your application's server token. Available at
                developer.uber.com.
            oauth2credential (OAuth2Credential)
                Access token and additional OAuth 2.0 credentials used
                to access protected resources.

        Raises
            UberIllegalState (APIError)
                Raised if there is an attempt to create session with
                both server token and access token.
        """
        if server_token and oauth2credential:
            message = (
                'Session cannot have both Server '
                'and OAuth 2.0 Credentials.'
            )
            raise UberIllegalState(message)

        if server_token is None and oauth2credential is None:
            message = (
                'Session must have either Server '
                'Token or OAuth 2.0 Credentials.'
            )
            raise UberIllegalState(message)

        if server_token:
            self.server_token = server_token
            self.token_type = auth.SERVER_TOKEN_TYPE
            self.oauth2credential = None

        elif oauth2credential:
            self.oauth2credential = oauth2credential
            self.token_type = auth.OAUTH_TOKEN_TYPE
            self.server_token = None


class OAuth2Credential(object):
    """A class to store OAuth 2.0 credentials.

    OAuth 2.0 credentials are used to properly construct requests
    to Uber and access protected resources. The class also stores
    app information (such as client_id) to refresh or request new
    access tokens if they expire or are revoked.
    """

    def __init__(
        self,
        client_id,
        access_token,
        expires_in_seconds,
        scopes,
        grant_type,
        redirect_url=None,
        client_secret=None,
        refresh_token=None,
    ):
        """Initialize an OAuth2Credential.

        Parameters
            client_id (str)
                Your app's Client ID.
            access_token (str)
                Access token received from OAuth 2.0 Authorization.
            expires_in_seconds (int)
                Seconds after initial grant when access token will expire.
            scopes (set)
                Set of permission scopes to request.
                (e.g. {'profile', 'history'}) Keep this list minimal so
                users feel safe granting your app access to their information.
            grant_type (str)
                Type of OAuth 2.0 Grant used to obtain access token.
                (e.g. 'authorization_code')
            redirect_url (str)
                The URL that the Uber server will redirect to.
            client_secret (str)
                Your app's Client Secret.
            refresh_token (str)
                Optional refresh token used to get a new access token.
                Only used for Authorization Code Grant.
        """
        self.client_id = client_id
        self.access_token = access_token
        self.expires_in_seconds = self._now() + int(expires_in_seconds)
        self.scopes = scopes
        self.grant_type = grant_type
        self.redirect_url = redirect_url
        self.client_secret = client_secret
        self.refresh_token = refresh_token

    @classmethod
    def make_from_response(
        cls,
        response,
        grant_type,
        client_id,
        client_secret=None,
        redirect_url=None,
    ):
        """Alternate constructor for OAuth2Credential().

        Create an OAuth2Credential from an HTTP Response.

        Parameters
            response (Response)
                HTTP Response containing OAuth 2.0 credentials.
            grant_type (str)
                Type of OAuth 2.0 Grant used to obtain access token.
                (e.g. 'authorization_code')
            client_id (str)
                Your app's Client ID.
            client_secret (str)
                Your app's Client Secret.
            redirect_url (str)
                The URL that the Uber server will redirect to.

        Returns
            (OAuth2Credential)

        Raises
            ClientError (APIError)
                Raised if the response is invalid.
        """
        if response.status_code != codes.ok:
            message = 'Error with Access Token Request: {}'
            message = message.format(response.reason)
            raise ClientError(response, message)

        response = response.json()

        # convert space delimited string to set
        scopes = response.get('scope')
        scopes_set = {scope for scope in scopes.split()}

        return cls(
            client_id=client_id,
            client_secret=client_secret,
            redirect_url=redirect_url,
            access_token=response.get('access_token'),
            expires_in_seconds=response.get('expires_in'),
            scopes=scopes_set,
            grant_type=grant_type,
            refresh_token=response.get('refresh_token', None),
        )

    def is_stale(self):
        """Check whether the session's current access token is about to expire.

        Returns
            (bool)
                True if access_token expires within threshold
        """
        expires_in_seconds = self.expires_in_seconds - self._now()
        return expires_in_seconds < EXPIRES_THRESHOLD_SECONDS

    def _now(self):
        return int(time())



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class APIError(Exception):
    """Parent class of all Uber API errors."""
    pass


class HTTPError(APIError):
    """Parent class of all HTTP errors."""

    def _adapt_response(self, response):
        """Convert error responses to standardized ErrorDetails."""
        if response.headers['content-type'] == 'application/json':
            body = response.json()
            status = response.status_code

            if body.get('errors'):
                return self._complex_response_to_error_adapter(body)

            elif body.get('code') and body.get('message'):
                return self._simple_response_to_error_adapter(status, body)

            elif body.get('error'):
                code = response.reason
                return self._message_to_error_adapter(status, code, body)

        raise UnknownHttpError(response)

    def _complex_response_to_error_adapter(self, body):
        """Convert a list of error responses."""
        meta = body.get('meta')
        errors = body.get('errors')
        e = []

        for error in errors:
            status = error['status']
            code = error['code']
            title = error['title']

            e.append(ErrorDetails(status, code, title))

        return e, meta

    def _simple_response_to_error_adapter(self, status, original_body):
        """Convert a single error response."""
        body = original_body.copy()
        code = body.pop('code')
        title = body.pop('message')
        meta = body    # save whatever is left in the response

        e = [ErrorDetails(status, code, title)]

        return e, meta

    def _message_to_error_adapter(self, status, code, original_body):
        """Convert single string message to error response."""
        body = original_body.copy()
        title = body.pop('error')
        meta = body    # save whatever is left in the response

        e = [ErrorDetails(status, code, title)]

        return e, meta


class ClientError(HTTPError):
    """Raise for 4XX Errors.

    Contains an array of ErrorDetails objects.
    """

    def __init__(self, response, message=None):
        """
        Parameters
            response
                The 4XX HTTP response.
            message
                An error message string. If one is not provided, the
                default message is used.
        """
        if not message:
            message = (
                'The request contains bad syntax or cannot be filled '
                'due to a fault from the client sending the request.'
            )

        super(ClientError, self).__init__(message)
        errors, meta = super(ClientError, self)._adapt_response(response)
        self.errors = errors
        self.meta = meta


class ServerError(HTTPError):
    """Raise for 5XX Errors.

    Contains a single ErrorDetails object.
    """

    def __init__(self, response, message=None):
        """
        Parameters
            response
                The 5XX HTTP response.
            message
                An error message string. If one is not provided, the
                default message is used.
        """
        if not message:
            message = (
                'The server encounter an error or is '
                'unable to process the request.'
            )

        super(ServerError, self).__init__(message)
        self.error, self.meta = self._adapt_response(response)

    def _adapt_response(self, response):
        """Convert various error responses to standardized ErrorDetails."""
        errors, meta = super(ServerError, self)._adapt_response(response)
        return errors[0], meta  # single error instead of array


class ErrorDetails(object):
    """Class to standardize all errors."""

    def __init__(self, status, code, title):
        self.status = status
        self.code = code
        self.title = title

    def __repr__(self):
        return "ErrorDetails: {} {} {}".format(
            str(self.status),
            str(self.code),
            str(self.title)
        )


class UnknownHttpError(APIError):
    """Throw when an unknown HTTP error occurs.

    Thrown when a previously unseen error is
    received and there is no standard schema to convert
    it into a well-formed HttpError.
    """

    def __init__(self, response):
        super(UnknownHttpError, self).__init__()
        self.response = response


class UberIllegalState(APIError):
    """Raise for Illegal State Errors.

    Thrown when the environment or class is not in an
    appropriate state for the requested operation.
    """
    pass



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Internal module for HTTP Requests and Responses."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from requests import Session
from string import ascii_letters
from string import digits

from uber_rides.errors import UberIllegalState
from uber_rides.utils import http
from uber_rides.utils.request import build_url
from uber_rides.utils.request import generate_data
from uber_rides.utils.request import generate_prepared_request


LIB_VERSION = '0.4.1'


class Response(object):
    """The response from an HTTP request."""

    def __init__(self, response):
        """Initialize a Response.

        Parameters
            response (requests.Response)
                The HTTP response from an API request.
        """
        self.status_code = response.status_code
        self.request = response.request
        self.headers = response.headers

        try:
            self.json = response.json()
        except:
            self.json = None


class Request(object):
    """Request containing information to send to server."""

    def __init__(
        self,
        auth_session,
        api_host,
        method,
        path,
        handlers=None,
        args=None,
    ):
        """Initialize a Request.

        Parameters
            auth_session (Session)
                Session object containing OAuth 2.0 credentials. Optional
                for any HTTP Requests that don't need access headers.
            api_host (str)
                Base URL of the Uber Server that handles API calls.
            method (str)
                HTTP Method (e.g. 'POST').
            path (str)
                The endpoint path. (e.g. 'v1.2/products')
            handlers (list[handler])
                Optional list of error handlers to attach to the request.
            args (dict)
                Optional dictionary of arguments to add to the request.
        """
        self.auth_session = auth_session
        self.api_host = api_host
        self.path = path
        self.method = method
        self.handlers = handlers or []
        self.args = args

    def _prepare(self):
        """Builds a URL and return a PreparedRequest.

        Returns
            (requests.PreparedRequest)

        Raises
            UberIllegalState (APIError)
        """
        if self.method not in http.ALLOWED_METHODS:
            raise UberIllegalState('Unsupported HTTP Method.')

        api_host = self.api_host
        headers = self._build_headers(self.method, self.auth_session)
        url = build_url(api_host, self.path)
        data, params = generate_data(self.method, self.args)

        return generate_prepared_request(
            self.method,
            url,
            headers,
            data,
            params,
            self.handlers,
        )

    def _send(self, prepared_request):
        """Send a PreparedRequest to the server.

        Parameters
            prepared_request (requests.PreparedRequest)

        Returns
            (Response)
                A Response object, whichcontains a server's
                response to an HTTP request.
        """
        session = Session()
        response = session.send(prepared_request)
        return Response(response)

    def execute(self):
        """Prepare and send the Request, return a Response.

        Returns
            (Response)
                The HTTP Response from an API Request
                to the server.

        Example
            request = Request(session, 'api.uber.com', 'GET', 'v1.2/profile')
            response = request.execute()
        """
        prepared_request = self._prepare()
        return self._send(prepared_request)

    def _build_headers(self, method, auth_session):
        """Create headers for the request.

        Parameters
            method (str)
                HTTP method (e.g. 'POST').
            auth_session (Session)
                The Session object containing OAuth 2.0 credentials.

        Returns
            headers (dict)
                Dictionary of access headers to attach to request.

        Raises
            UberIllegalState (ApiError)
                Raised if headers are invalid.
        """
        token_type = auth_session.token_type

        if auth_session.server_token:
            token = auth_session.server_token
        else:
            token = auth_session.oauth2credential.access_token

        if not self._authorization_headers_valid(token_type, token):
            message = 'Invalid token_type or token.'
            raise UberIllegalState(message)

        headers = {
            'Authorization': ' '.join([token_type, token]),
            'X-Uber-User-Agent': 'Python Rides SDK v{}'.format(LIB_VERSION),
        }

        if method in http.BODY_METHODS:
            headers.update(http.DEFAULT_CONTENT_HEADERS)

        return headers

    def _authorization_headers_valid(self, token_type, token):
        """Verify authorization headers for a request.

        Parameters
            token_type (str)
                Type of token to access resources.
            token (str)
                Server Token or OAuth 2.0 Access Token.

        Returns
            (bool)
                True iff token_type and token are valid.
        """
        if token_type not in http.VALID_TOKEN_TYPES:
            return False

        allowed_chars = ascii_letters + digits + '.' + '_' + '-'

        # True if token only contains allowed_chars
        return all(characters in allowed_chars for characters in token)



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pytest import fixture
from pytest import raises

try:
    from urllib.parse import quote
    from urllib.parse import parse_qs
    from urllib.parse import urlparse
except ImportError:
    from urllib import quote
    from urlparse import parse_qs
    from urlparse import urlparse

from tests.vcr_config import uber_vcr
from uber_rides.auth import AuthorizationCodeGrant
from uber_rides.auth import ClientCredentialGrant
from uber_rides.auth import ImplicitGrant
from uber_rides.auth import refresh_access_token
from uber_rides.errors import UberIllegalState
from uber_rides.session import OAuth2Credential
from uber_rides.utils import auth


# replace these with valid tokens and credentials to rerecord fixtures
CLIENT_ID = 'xxx'
CLIENT_SECRET = 'xxx'
AUTHORIZATION_CODE = 'xxx'
ACCESS_TOKEN = 'xxx'
REFRESH_TOKEN = 'xxx'
REFRESHED_ACCESS_TOKEN = 'xxx'
STATE_TOKEN = 'rides-sdk'

REDIRECT_URL = 'https://uberapitester.com/api/v1/uber/oauth'

SCOPES = {'profile', 'history'}
CLIENT_CREDENTIALS_SCOPES = {'partner.referrals'}

EXPIRES_IN_SECONDS = 3000

EXPECTED_QUERYSTRING = (
    'scope={}&state={}&redirect_uri={}'
    '&response_type={}&client_id={}'
)


@fixture
def auth_code_grant():
    return AuthorizationCodeGrant(
        client_id=CLIENT_ID,
        scopes=SCOPES,
        client_secret=CLIENT_SECRET,
        redirect_url=REDIRECT_URL,
    )


@fixture
def auth_code_oauth2credential():
    return OAuth2Credential(
        client_id=CLIENT_ID,
        client_secret=CLIENT_SECRET,
        redirect_url=REDIRECT_URL,
        access_token=ACCESS_TOKEN,
        expires_in_seconds=EXPIRES_IN_SECONDS,
        scopes=SCOPES,
        grant_type=auth.AUTHORIZATION_CODE_GRANT,
        refresh_token=REFRESH_TOKEN,
    )


@fixture
def implicit_grant():
    return ImplicitGrant(
        client_id=CLIENT_ID,
        scopes=SCOPES,
        redirect_url=REDIRECT_URL,
    )


@fixture
def implicit_oauth2credential():
    return OAuth2Credential(
        client_id=CLIENT_ID,
        client_secret=None,
        redirect_url=REDIRECT_URL,
        access_token=ACCESS_TOKEN,
        expires_in_seconds=EXPIRES_IN_SECONDS,
        scopes=SCOPES,
        grant_type=auth.IMPLICIT_GRANT,
        refresh_token=None,
    )


@fixture
def client_credential_grant():
    return ClientCredentialGrant(
        client_id=CLIENT_ID,
        scopes=CLIENT_CREDENTIALS_SCOPES,
        client_secret=CLIENT_SECRET,
    )


@fixture
def client_credential_oauth2credential():
    return OAuth2Credential(
        client_id=CLIENT_ID,
        client_secret=CLIENT_SECRET,
        redirect_url=None,
        access_token=ACCESS_TOKEN,
        expires_in_seconds=EXPIRES_IN_SECONDS,
        scopes=CLIENT_CREDENTIALS_SCOPES,
        grant_type=auth.CLIENT_CREDENTIAL_GRANT,
        refresh_token=None,
    )


def test_auth_code_grant_get_authorization_url(auth_code_grant):
    """Test to get authorization url for authorization code grant."""
    url = auth_code_grant.get_authorization_url()
    parsed_url = urlparse(url)
    assert parsed_url.netloc == auth.AUTH_HOST
    assert parsed_url.path.strip('/') == auth.AUTHORIZE_PATH

    querystring = parsed_url.query
    queryparams = parse_qs(querystring)
    expected_query = EXPECTED_QUERYSTRING.format(
        '+'.join(SCOPES),
        auth_code_grant.state_token,
        quote(REDIRECT_URL, safe=''),
        auth.CODE_RESPONSE_TYPE,
        CLIENT_ID,
    )

    assert expected_query in querystring
    assert queryparams.get('scope')[0] == ' '.join(SCOPES)
    assert queryparams.get('redirect_uri')[0] == REDIRECT_URL
    assert queryparams.get('response_type')[0] == auth.CODE_RESPONSE_TYPE
    assert queryparams.get('client_id')[0] == CLIENT_ID
    assert queryparams.get('state')[0] == auth_code_grant.state_token


@uber_vcr.use_cassette()
def test_auth_code_get_session(auth_code_grant):
    """Test to get OAuth 2.0 session for authorization code grant."""
    auth_code_grant.state_token = STATE_TOKEN
    redirect_url = '{}?state={}&code={}'
    redirect_url = redirect_url.format(
        REDIRECT_URL,
        STATE_TOKEN,
        AUTHORIZATION_CODE,
    )
    session = auth_code_grant.get_session(redirect_url)

    assert session.server_token is None
    assert session.token_type == auth.OAUTH_TOKEN_TYPE
    credential = session.oauth2credential
    assert credential.access_token == ACCESS_TOKEN
    assert credential.refresh_token == REFRESH_TOKEN
    assert credential.scopes == SCOPES
    assert credential.grant_type == auth.AUTHORIZATION_CODE_GRANT
    assert credential.client_id == CLIENT_ID
    assert credential.client_secret == CLIENT_SECRET
    assert credential.redirect_url == REDIRECT_URL


@uber_vcr.use_cassette()
def test_refresh_auth_code_access_token(auth_code_oauth2credential):
    """Test to refresh OAuth 2.0 session for authorization code grant."""
    session = refresh_access_token(auth_code_oauth2credential)

    assert session.server_token is None
    assert session.token_type == auth.OAUTH_TOKEN_TYPE
    credential = session.oauth2credential
    assert credential.access_token == REFRESHED_ACCESS_TOKEN
    assert credential.refresh_token
    assert credential.scopes == SCOPES
    assert credential.grant_type == auth.AUTHORIZATION_CODE_GRANT
    assert credential.client_id == CLIENT_ID
    assert credential.client_secret == CLIENT_SECRET
    assert credential.redirect_url == REDIRECT_URL


def test_implicit_grant_get_authorization_url(implicit_grant):
    """Test to get authorization url for implicit grant."""
    url = implicit_grant.get_authorization_url()
    parsed_url = urlparse(url)
    assert parsed_url.netloc == auth.AUTH_HOST
    assert parsed_url.path.strip('/') == auth.AUTHORIZE_PATH

    querystring = parsed_url.query
    queryparams = parse_qs(querystring)
    expected_query = EXPECTED_QUERYSTRING.format(
        '+'.join(SCOPES),
        None,
        quote(REDIRECT_URL, safe=''),
        auth.TOKEN_RESPONSE_TYPE,
        CLIENT_ID,
    )

    assert expected_query in querystring
    assert queryparams.get('scope')[0] == ' '.join(SCOPES)
    assert queryparams.get('redirect_uri')[0] == REDIRECT_URL
    assert queryparams.get('response_type')[0] == auth.TOKEN_RESPONSE_TYPE
    assert queryparams.get('client_id')[0] == CLIENT_ID


def test_implicit_grant_get_session(implicit_grant):
    """Test to get OAuth 2.0 session from URL for implicit grant."""
    redirect_url = (
        '{}#access_token={}&token_type=Bearer&state=None'
        '&expires_in=2592000&scope=profile+history'
    )
    redirect_url = redirect_url.format(REDIRECT_URL, ACCESS_TOKEN)
    session = implicit_grant.get_session(redirect_url)

    assert session.server_token is None
    assert session.token_type == auth.OAUTH_TOKEN_TYPE
    credential = session.oauth2credential
    assert credential.access_token == ACCESS_TOKEN
    assert credential.scopes == SCOPES
    assert credential.grant_type == auth.IMPLICIT_GRANT
    assert credential.client_id == CLIENT_ID
    assert credential.redirect_url == REDIRECT_URL
    assert credential.client_secret is None
    assert credential.refresh_token is None


def test_refresh_implicit_access_token(implicit_oauth2credential):
    """Test that refreshing credentials for implicit grant raises error."""
    with raises(UberIllegalState) as error:
        refresh_access_token(implicit_oauth2credential)

    assert 'Grant Type does not support Refresh Tokens' in str(error.value)


@uber_vcr.use_cassette()
def test_client_credential_get_session(client_credential_grant):
    """Test to get OAuth 2.0 session for client credential grant."""
    session = client_credential_grant.get_session()

    assert session.server_token is None
    assert session.token_type == auth.OAUTH_TOKEN_TYPE
    credential = session.oauth2credential
    assert credential.access_token == ACCESS_TOKEN
    assert credential.scopes == CLIENT_CREDENTIALS_SCOPES
    assert credential.grant_type == auth.CLIENT_CREDENTIAL_GRANT
    assert credential.client_id == CLIENT_ID
    assert credential.client_secret == CLIENT_SECRET
    assert credential.refresh_token is None
    assert credential.redirect_url is None


@uber_vcr.use_cassette()
def test_refresh_client_credential_access_token(
    client_credential_oauth2credential,
):
    """Test to refresh OAuth 2.0 session for client credential grant."""
    session = refresh_access_token(client_credential_oauth2credential)

    assert session.server_token is None
    assert session.token_type == auth.OAUTH_TOKEN_TYPE
    credential = session.oauth2credential
    assert credential.access_token == REFRESHED_ACCESS_TOKEN
    assert credential.scopes == CLIENT_CREDENTIALS_SCOPES
    assert credential.grant_type == auth.CLIENT_CREDENTIAL_GRANT
    assert credential.client_id == CLIENT_ID
    assert credential.client_secret == CLIENT_SECRET
    assert credential.redirect_url is None
    assert credential.refresh_token is None



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Use an UberRidesClient to request and complete a ride.

This example demonstrates how to use an UberRidesClient to request a ride
under surge. After successfully requesting a ride, it updates the
ride status to 'completed' and deactivates surge.

To run this example:

    (1) Run `python authorize_user.py` to get OAuth 2.0 Credentials
    (2) Run `python request_ride.py`
    (3) The UberRidesClient will make API calls and print the
        results to your terminal.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

try:
    from urllib.parse import parse_qs
    from urllib.parse import urlparse
except ImportError:
    from urlparse import parse_qs
    from urlparse import urlparse

from builtins import input

from example.utils import create_uber_client
from example.utils import fail_print
from example.utils import import_oauth2_credentials
from example.utils import paragraph_print
from example.utils import response_print
from example.utils import success_print

from uber_rides.client import SurgeError
from uber_rides.errors import ClientError
from uber_rides.errors import ServerError

# uber pool
UFP_PRODUCT_ID = '26546650-e557-4a7b-86e7-6a3942445247'

# uber black
SURGE_PRODUCT_ID = 'd4abaae7-f4d6-4152-91cc-77523e8165a4'

# California Academy of Sciences
START_LAT = 37.770
START_LNG = -122.466

# Uber HQ
END_LAT = 37.7752315
END_LNG = -122.418075


def estimate_ride(api_client):
    """Use an UberRidesClient to fetch a ride estimate and print the results.

    Parameters
        api_client (UberRidesClient)
            An authorized UberRidesClient with 'request' scope.
    """
    try:
        estimate = api_client.estimate_ride(
            product_id=SURGE_PRODUCT_ID,
            start_latitude=START_LAT,
            start_longitude=START_LNG,
            end_latitude=END_LAT,
            end_longitude=END_LNG,
            seat_count=2
        )

    except (ClientError, ServerError) as error:
        fail_print(error)

    else:
        success_print(estimate.json)


def update_surge(api_client, surge_multiplier):
    """Use an UberRidesClient to update surge and print the results.

    Parameters
        api_client (UberRidesClient)
            An authorized UberRidesClient with 'request' scope.
        surge_mutliplier (float)
            The surge multiple for a sandbox product. A multiplier greater than
            or equal to 2.0 will require the two stage confirmation screen.
    """
    try:
        update_surge = api_client.update_sandbox_product(
            SURGE_PRODUCT_ID,
            surge_multiplier=surge_multiplier,
        )

    except (ClientError, ServerError) as error:
        fail_print(error)

    else:
        success_print(update_surge.status_code)


def update_ride(api_client, ride_status, ride_id):
    """Use an UberRidesClient to update ride status and print the results.

    Parameters
        api_client (UberRidesClient)
            An authorized UberRidesClient with 'request' scope.
        ride_status (str)
            New ride status to update to.
        ride_id (str)
            Unique identifier for ride to update.
    """
    try:
        update_product = api_client.update_sandbox_ride(ride_id, ride_status)

    except (ClientError, ServerError) as error:
        fail_print(error)

    else:
        message = '{} New status: {}'
        message = message.format(update_product.status_code, ride_status)
        success_print(message)


def request_ufp_ride(api_client):
    """Use an UberRidesClient to request a ride and print the results.

    Parameters
        api_client (UberRidesClient)
            An authorized UberRidesClient with 'request' scope.

    Returns
        The unique ID of the requested ride.
    """
    try:

        estimate = api_client.estimate_ride(
            product_id=UFP_PRODUCT_ID,
            start_latitude=START_LAT,
            start_longitude=START_LNG,
            end_latitude=END_LAT,
            end_longitude=END_LNG,
            seat_count=2
        )
        fare = estimate.json.get('fare')

        request = api_client.request_ride(
            product_id=UFP_PRODUCT_ID,
            start_latitude=START_LAT,
            start_longitude=START_LNG,
            end_latitude=END_LAT,
            end_longitude=END_LNG,
            seat_count=2,
            fare_id=fare['fare_id']
        )

    except (ClientError, ServerError) as error:
        fail_print(error)
        return

    else:
        success_print(estimate.json)
        success_print(request.json)
        return request.json.get('request_id')


def request_surge_ride(api_client, surge_confirmation_id=None):
    """Use an UberRidesClient to request a ride and print the results.

    If the product has a surge_multiple greater than or equal to 2.0,
    a SurgeError is raised. Confirm surge by visiting the
    surge_confirmation_url and automatically try the request again.

    Parameters
        api_client (UberRidesClient)
            An authorized UberRidesClient with 'request' scope.
        surge_confirmation_id (string)
            Unique identifer received after confirming surge.

    Returns
        The unique ID of the requested ride.
    """
    try:
        request = api_client.request_ride(
            product_id=SURGE_PRODUCT_ID,
            start_latitude=START_LAT,
            start_longitude=START_LNG,
            end_latitude=END_LAT,
            end_longitude=END_LNG,
            surge_confirmation_id=surge_confirmation_id,
            seat_count=2
        )

    except SurgeError as e:
        surge_message = 'Confirm surge by visiting: \n{}\n'
        surge_message = surge_message.format(e.surge_confirmation_href)
        response_print(surge_message)

        confirm_url = 'Copy the URL you are redirected to and paste here: \n'
        result = input(confirm_url).strip()

        querystring = urlparse(result).query
        query_params = parse_qs(querystring)
        surge_id = query_params.get('surge_confirmation_id')[0]

        # automatically try request again
        return request_surge_ride(api_client, surge_id)

    except (ClientError, ServerError) as error:
        fail_print(error)
        return

    else:
        success_print(request.json)
        return request.json.get('request_id')


def get_ride_details(api_client, ride_id):
    """Use an UberRidesClient to get ride details and print the results.

    Parameters
        api_client (UberRidesClient)
            An authorized UberRidesClient with 'request' scope.
        ride_id (str)
            Unique ride identifier.
    """
    try:
        ride_details = api_client.get_ride_details(ride_id)

    except (ClientError, ServerError) as error:
        fail_print(error)

    else:
        success_print(ride_details.json)


if __name__ == '__main__':
    """Run the example.

    Create an UberRidesClient from OAuth 2.0 Credentials, update a sandbox
    product's surge, request and complete a ride.
    """
    credentials = import_oauth2_credentials()
    api_client = create_uber_client(credentials)

    # ride request with upfront pricing flow

    paragraph_print("Request a ride with upfront pricing product.")
    ride_id = request_ufp_ride(api_client)

    paragraph_print("Update ride status to accepted.")
    update_ride(api_client, 'accepted', ride_id)

    paragraph_print("Updated ride details.")
    get_ride_details(api_client, ride_id)
    update_ride(api_client, 'in_progress', ride_id)

    paragraph_print("Updated ride details.")
    get_ride_details(api_client, ride_id)

    paragraph_print("Update ride status to completed.")
    update_ride(api_client, 'completed', ride_id)

    paragraph_print("Updated ride details.")
    get_ride_details(api_client, ride_id)

    # ride request with surge flow

    paragraph_print("Ride estimates before surge.")
    estimate_ride(api_client)

    paragraph_print("Activate surge.")
    update_surge(api_client, 2.0)

    paragraph_print("Ride estimates after surge.")
    estimate_ride(api_client)

    paragraph_print("Request a ride with surging product.")
    ride_id = request_surge_ride(api_client)

    paragraph_print("Update ride status to accepted.")
    update_ride(api_client, 'accepted', ride_id)

    paragraph_print("Updated ride details.")
    get_ride_details(api_client, ride_id)
    update_ride(api_client, 'in_progress', ride_id)

    paragraph_print("Updated ride details.")
    get_ride_details(api_client, ride_id)

    paragraph_print("Update ride status to completed.")
    update_ride(api_client, 'completed', ride_id)

    paragraph_print("Updated ride details.")
    get_ride_details(api_client, ride_id)

    paragraph_print("Deactivate surge.")
    update_surge(api_client, 1.0)



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from mock import Mock
from pytest import fixture
from pytest import raises
from requests import codes

from tests.vcr_config import uber_vcr
from uber_rides.client import SurgeError
from uber_rides.client import UberRidesClient
from uber_rides.errors import ErrorDetails
from uber_rides.session import OAuth2Credential
from uber_rides.session import Session
from uber_rides.utils import auth
from uber_rides.utils import http


# replace these with valid tokens and credentials to rerecord fixtures
CLIENT_ID = 'xxx'
CLIENT_SECRET = 'xxx'
SERVER_TOKEN = 'xxx'

ACCESS_TOKEN = 'xxx'
REFRESH_TOKEN = 'xxx'
REDIRECT_URL = 'https://uberapitester.com/api/v1/uber/oauth'

RIDER_SCOPES = {
          'profile',
          'places',
          'request',
          'request_receipt',
          'all_trips',
          'history'
          }

DRIVER_SCOPES = {
          'partner.accounts',
          'partner.trips',
          'partner.payments'
          }

# replace these with valid identifiers to rerecord request-related fixtures
RIDE_ID = '0aec0061-1e20-4239-a0b7-78328e9afec8'
SURGE_ID = 'hsg2k38b'

# addresses to test places
HOME_ADDRESS = '555 Market Street, SF'
FULL_HOME_ADDRESS = '555 Market St, San Francisco, CA 94105, USA'
WORK_ADDRESS = '1455 Market Street, SF'
FULL_WORK_ADDRESS = '1455 Market St, San Francisco, CA 94103, USA'

# ride details
EXPIRES_IN_SECONDS = 3000
START_LAT = 37.7899886
START_LNG = -122.4021253
END_LAT = 37.775232
END_LNG = -122.4197513
UPDATE_LAT = 37.775234
UPDATE_LNG = -122.4197515
NON_UFP_PRODUCT_ID = '3ab64887-4842-4c8e-9780-ccecd3a0391d'
UFP_PRODUCT_ID = '821415d8-3bd5-4e27-9604-194e4359a449'
UFP_FARE_ID = '7dad38f13eab3124621d16604c35fb26e3' \
              '0395e76937c507565fb1b4aa4a8264'
UFP_SHARED_PRODUCT_ID = '26546650-e557-4a7b-86e7-6a3942445247'
UFP_SHARED_SEAT_COUNT = 2
UFP_SHARED_FARE_ID = 'd30e732b8bba22c9cdc10513ee86380087cb'\
                     '4a6f89e37ad21ba2a39f3a1ba960'
PRODUCTS_AVAILABLE = 8

NO_DESTINATION_PRODUCT_ID = '2541f77c-920a-45c4-8bf3-603ecd625195'

NON_UFP_SURGE_PRODUCT_ID = 'd5ef01d9-7d54-413e-b265-425948d06e92'
NON_UFP_SURGE_START_LAT = -22.9674153
NON_UFP_SURGE_START_LNG = -43.1801978
NON_UFP_SURGE_END_LAT = -22.9883286
NON_UFP_SURGE_END_LNG = -43.1925661
SURGE_HREF = 'api.uber.com/v1.2/surge-confirmations/{}'

EXPECTED_PRODUCT_KEYS = set([
    'capacity',
    'description',
    'image',
    'display_name',
    'product_id',
    'shared',
    'short_description',
    'product_group',
    'cash_enabled',
    'upfront_fare_enabled'
])

EXPECTED_TIME_KEYS = set([
    'localized_display_name',
    'estimate',
    'display_name',
    'product_id'
])

EXPECTED_PRICE_KEYS = set([
    'high_estimate',
    'low_estimate',
    'currency_code',
    'localized_display_name',
    'estimate',
    'distance',
    'duration',
    'display_name',
    'product_id'
])

EXPECTED_PROMOTION_KEYS = set([
    'display_text',
    'localized_value',
    'type'
])

EXPECTED_ACTIVITY_KEYS = set([
    'status',
    'distance',
    'start_time',
    'start_city',
    'end_time',
    'request_id',
    'product_id'
])

EXPECTED_PROFILE_KEYS = set([
    'picture',
    'first_name',
    'last_name',
    'uuid',
    'email',
    'promo_code'
])

EXPECTED_ESTIMATE_RIDE_FARE_KEYS = set([
    'value',
    'fare_id',
    'expires_at',
    'currency_code',
    'display'
])

EXPECTED_ESTIMATE_RIDE_TRIP_KEYS = set([
    'distance_unit',
    'duration_estimate',
    'distance_estimate'
])

EXPECTED_ESTIMATE_SHARED_RIDE_TRIP_KEYS = set([
    'distance_unit',
    'duration_estimate',
    'distance_estimate'
])

EXPECTED_RIDE_DETAILS_KEYS = set([
    'status',
    'request_id',
    'product_id',
    'driver',
    'pickup',
    'destination',
    'location',
    'vehicle',
    'shared'
])

EXPECTED_SHARED_RIDE_DETAILS_KEYS = set([
    'status',
    'request_id',
    'product_id',
    'driver',
    'pickup',
    'destination',
    'location',
    'vehicle',
    'shared'
])

EXPECTED_RIDE_MAP_KEYS = set([
    'href',
    'request_id'
])

EXPECTED_INDIVIDUAL_CHARGE_KEYS = set([
    'amount',
    'type',
    'name'
])

EXPECTED_RECEIPT_KEYS = set([
    'distance',
    'charge_adjustments',
    'total_owed',
    'total_fare',
    'total_charged',
    'distance_label',
    'request_id',
    'duration',
    'subtotal'
])

EXPECTED_PLACE_KEYS = set(['address'])

EXPECTED_PAYMENT_KEYS = set([
    'payment_method_id',
    'type',
    'description'
])

EXPECTED_DRIVER_PROFILE_KEYS = set([
    'driver_id',
    'activation_status',
    'first_name',
    'last_name',
    'driver_id',
    'email',
    'phone_number',
    'picture',
    'promo_code',
    'rating'
])

EXPECTED_DRIVER_TRIPS_KEYS = set([
    'dropoff',
    'distance',
    'status_changes',
    'pickup',
    'driver_id',
    'status',
    'trip_id'
])

EXPECTED_DRIVER_PAYMENTS_KEYS = set([
    'category',
    'breakdown',
    'rider_fees',
    'event_time',
    'trip_id',
    'payment_id',
    'amount',
    'driver_id',
    'partner_id',
    'currency_code'
])


@fixture
def rider_oauth2credential():
    """Create OAuth2Credential class to hold access token information."""
    return OAuth2Credential(
        client_id=CLIENT_ID,
        redirect_url=REDIRECT_URL,
        access_token=ACCESS_TOKEN,
        expires_in_seconds=EXPIRES_IN_SECONDS,
        scopes=RIDER_SCOPES,
        grant_type=auth.AUTHORIZATION_CODE_GRANT,
        client_secret=CLIENT_SECRET,
        refresh_token=REFRESH_TOKEN,
    )


@fixture
def driver_oauth2credential():
    """Create OAuth2Credential class to hold access token information."""
    return OAuth2Credential(
        client_id=CLIENT_ID,
        redirect_url=REDIRECT_URL,
        access_token=ACCESS_TOKEN,
        expires_in_seconds=EXPIRES_IN_SECONDS,
        scopes=DRIVER_SCOPES,
        grant_type=auth.AUTHORIZATION_CODE_GRANT,
        client_secret=CLIENT_SECRET,
        refresh_token=REFRESH_TOKEN,
    )


@fixture
def authorized_rider_sandbox_client(rider_oauth2credential):
    """Create an UberRidesClient in Sandbox Mode with OAuth 2.0 Credentials."""
    session = Session(oauth2credential=rider_oauth2credential)
    return UberRidesClient(session, sandbox_mode=True)


@fixture
def authorized_rider_production_client(rider_oauth2credential):
    """Create an UberRidesClient in Production with OAuth 2.0 Credentials."""
    session = Session(oauth2credential=rider_oauth2credential)
    return UberRidesClient(session)


@fixture
def authorized_driver_sandbox_client(driver_oauth2credential):
    """Create an UberRidesClient in Sandbox Mode with OAuth 2.0 Credentials."""
    session = Session(oauth2credential=driver_oauth2credential)
    return UberRidesClient(session, sandbox_mode=True)


@fixture
def authorized_driver_production_client(driver_oauth2credential):
    """Create an UberRidesClient in Production with OAuth 2.0 Credentials."""
    session = Session(oauth2credential=driver_oauth2credential)
    return UberRidesClient(session)


@fixture
def server_token_client():
    """Create an UberRidesClient with Server Token."""
    session = Session(server_token=SERVER_TOKEN)
    return UberRidesClient(session)


@fixture
def http_surge_error():
    code = 'surge'

    mock_error = Mock(
        status_code=http.STATUS_CONFLICT,
        headers=http.DEFAULT_CONTENT_HEADERS,
    )

    error_response = {
        'meta': {
            'surge_confirmation': {
                'href': SURGE_HREF.format(SURGE_ID),
                'surge_confirmation_id': SURGE_ID,
            },
        },
        'errors': [
            {
                'status': http.STATUS_CONFLICT,
                'code': code,
                'title': http.ERROR_CODE_DESCRIPTION_DICT[code],
            },
        ],
    }

    mock_error.json = Mock(return_value=error_response)
    return mock_error


@uber_vcr.use_cassette()
def test_get_products(authorized_rider_sandbox_client, server_token_client):
    """Test to fetch products with access token and server token."""
    clients = [authorized_rider_sandbox_client, server_token_client]

    for client in clients:
        response = client.get_products(START_LAT, START_LNG)
        assert response.status_code == codes.ok

        # assert response looks like products information
        response = response.json
        products = response.get('products')
        assert len(products) >= PRODUCTS_AVAILABLE
        assert isinstance(products, list)

        for product in products:
            assert EXPECTED_PRODUCT_KEYS.issubset(product)


@uber_vcr.use_cassette()
def test_get_single_product(authorized_rider_sandbox_client,
                            server_token_client
                            ):
    """Test fetch product by ID with access token and server token."""
    clients = [authorized_rider_sandbox_client, server_token_client]

    for client in clients:
        response = client.get_product(UFP_PRODUCT_ID)
        assert response.status_code == codes.ok

        # assert response looks like single product information
        response = response.json
        assert 'products' not in response
        assert EXPECTED_PRODUCT_KEYS.issubset(response)


@uber_vcr.use_cassette()
def test_get_price_estimates(authorized_rider_sandbox_client,
                             server_token_client
                             ):
    """Test to fetch price estimates with access token and server token."""
    clients = [authorized_rider_sandbox_client, server_token_client]

    for client in clients:
        response = client.get_price_estimates(
            START_LAT,
            START_LNG,
            END_LAT,
            END_LNG,
            UFP_SHARED_SEAT_COUNT,
        )
        assert response.status_code == codes.ok

        # assert response looks like price estimates
        response = response.json
        prices = response.get('prices')
        assert len(prices) >= PRODUCTS_AVAILABLE
        assert isinstance(prices, list)

        for price in prices:
            assert EXPECTED_PRICE_KEYS.issubset(price)


@uber_vcr.use_cassette()
def test_get_pickup_time_estimates(
    authorized_rider_sandbox_client,
    server_token_client,
):
    """Test to fetch time estimates with access token and server token."""
    clients = [authorized_rider_sandbox_client, server_token_client]

    for client in clients:
        response = client.get_pickup_time_estimates(
            START_LAT,
            START_LNG,
        )
        assert response.status_code == codes.ok

        # assert response looks like pickup time estimates
        response = response.json
        times = response.get('times')
        assert len(times) >= PRODUCTS_AVAILABLE
        assert isinstance(times, list)

        for pickup_time in times:
            assert EXPECTED_TIME_KEYS.issubset(pickup_time)


@uber_vcr.use_cassette()
def test_get_promotions(authorized_rider_sandbox_client, server_token_client):
    """Test to fetch promotions with access token and server token."""
    clients = [authorized_rider_sandbox_client, server_token_client]

    for client in clients:
        response = client.get_promotions(
            START_LAT,
            START_LNG,
            END_LAT,
            END_LNG,
        )
        assert response.status_code == codes.ok

        # assert response looks like promotions
        response = response.json
        assert EXPECTED_PROMOTION_KEYS.issubset(response)


@uber_vcr.use_cassette()
def test_get_user_activity(authorized_rider_sandbox_client):
    """Test to fetch user activity with access token."""
    response = authorized_rider_sandbox_client.get_user_activity()
    assert response.status_code == codes.ok

    # assert response looks like user activity
    response = response.json
    history = response.get('history')
    assert len(history) == 5
    assert isinstance(history, list)

    for activity in history:
        assert EXPECTED_ACTIVITY_KEYS.issubset(activity)


@uber_vcr.use_cassette()
def test_get_user_profile(authorized_rider_sandbox_client):
    """Test to fetch user profile with access token."""
    response = authorized_rider_sandbox_client.get_user_profile()
    assert response.status_code == codes.ok

    # assert response looks like user profile
    response = response.json
    assert EXPECTED_PROFILE_KEYS.issubset(response)


@uber_vcr.use_cassette()
def test_estimate_shared_ride(authorized_rider_sandbox_client):
    """Test to estimate a shared ride."""
    try:
        response = authorized_rider_sandbox_client.estimate_ride(
            product_id=UFP_SHARED_PRODUCT_ID,
            seat_count=UFP_SHARED_SEAT_COUNT,
            start_latitude=START_LAT,
            start_longitude=START_LNG,
            end_latitude=END_LAT,
            end_longitude=END_LNG,
        )
    except Exception as e:
        print(e.errors[0].__dict__)
    assert response.status_code == codes.ok

    # assert response looks like price and time estimates
    response = response.json
    fare = response.get('fare')
    assert EXPECTED_ESTIMATE_RIDE_FARE_KEYS.issubset(fare)
    trip = response.get('trip')
    assert EXPECTED_ESTIMATE_SHARED_RIDE_TRIP_KEYS.issubset(trip)


@uber_vcr.use_cassette()
def test_estimate_ride(authorized_rider_sandbox_client):
    """Test to fetch ride estimate with access token."""
    response = authorized_rider_sandbox_client.estimate_ride(
        product_id=UFP_PRODUCT_ID,
        start_latitude=START_LAT,
        start_longitude=START_LNG,
        end_latitude=END_LAT,
        end_longitude=END_LNG,
    )
    assert response.status_code == codes.ok

    # assert response looks like price and time estimates
    response = response.json
    fare = response.get('fare')
    assert EXPECTED_ESTIMATE_RIDE_FARE_KEYS.issubset(fare)
    trip = response.get('trip')
    assert EXPECTED_ESTIMATE_RIDE_TRIP_KEYS.issubset(trip)


@uber_vcr.use_cassette()
def test_estimate_ride_with_places(authorized_rider_sandbox_client):
    """Test to fetch ride estimate with place ids."""
    response = authorized_rider_sandbox_client.estimate_ride(
        product_id=UFP_PRODUCT_ID,
        start_place_id='home',
        end_place_id='work',
    )
    assert response.status_code == codes.ok

    # assert response looks like price and time estimates
    response = response.json
    fare = response.get('fare')
    assert EXPECTED_ESTIMATE_RIDE_FARE_KEYS.issubset(fare)
    trip = response.get('trip')
    assert EXPECTED_ESTIMATE_RIDE_TRIP_KEYS.issubset(trip)


@uber_vcr.use_cassette()
def test_request_ride(authorized_rider_sandbox_client):
    """Test to request ride with access token."""
    response = authorized_rider_sandbox_client.request_ride(
        product_id=UFP_PRODUCT_ID,
        fare_id=UFP_FARE_ID,
        start_latitude=START_LAT,
        start_longitude=START_LNG,
        end_latitude=END_LAT,
        end_longitude=END_LNG,
    )
    assert response.status_code == codes.accepted

    # assert response looks like ride details
    response = response.json
    assert EXPECTED_RIDE_DETAILS_KEYS.issubset(response)


@uber_vcr.use_cassette()
def test_request_shared_ride(authorized_rider_sandbox_client):
    """Test to request shared ride with access token."""
    try:
        response = authorized_rider_sandbox_client.request_ride(
            product_id=UFP_SHARED_PRODUCT_ID,
            fare_id=UFP_SHARED_FARE_ID,
            seat_count=UFP_SHARED_SEAT_COUNT,
            start_latitude=START_LAT,
            start_longitude=START_LNG,
            end_latitude=END_LAT,
            end_longitude=END_LNG,
        )
    except Exception as e:
        print(e)
        print(e.errors[0].__dict__)
    assert response.status_code == codes.accepted


@uber_vcr.use_cassette()
def test_request_ride_with_no_default_product(authorized_rider_sandbox_client):
    """Test to request ride with no default product."""
    response = authorized_rider_sandbox_client.request_ride(
        fare_id=UFP_FARE_ID,
        start_latitude=START_LAT,
        start_longitude=START_LNG,
        end_latitude=END_LAT,
        end_longitude=END_LNG,
    )
    assert response.status_code == codes.accepted

    # assert response looks like ride details
    response = response.json
    assert EXPECTED_RIDE_DETAILS_KEYS.issubset(response)


@uber_vcr.use_cassette()
def test_request_ride_with_places(authorized_rider_sandbox_client):
    """Test to request ride with place ids."""
    response = authorized_rider_sandbox_client.request_ride(
        product_id=UFP_PRODUCT_ID,
        fare_id=UFP_FARE_ID,
        start_place_id='home',
        end_place_id='work',
    )
    assert response.status_code == codes.accepted

    # assert response looks like ride details
    response = response.json
    assert EXPECTED_RIDE_DETAILS_KEYS.issubset(response)


@uber_vcr.use_cassette()
def test_get_ride_details(authorized_rider_sandbox_client):
    """Test to fetch ride details with access token."""
    response = authorized_rider_sandbox_client.get_ride_details(RIDE_ID)
    assert response.status_code == codes.ok

    # assert response looks like ride details
    response = response.json
    assert EXPECTED_RIDE_DETAILS_KEYS.issubset(response)
    assert response.get('status') == 'processing'


@uber_vcr.use_cassette()
def test_get_current_ride_details(authorized_rider_sandbox_client):
    """Test to fetch current ride details with access token."""
    response = authorized_rider_sandbox_client.get_current_ride_details()
    assert response.status_code == codes.ok

    # assert response looks like ride details
    response = response.json
    assert EXPECTED_RIDE_DETAILS_KEYS.issubset(response)
    assert response.get('status') == 'processing'


@uber_vcr.use_cassette()
def test_get_current_shared_ride_details(authorized_rider_sandbox_client):
    """Test to fetch current shared ride details with access token."""
    response = authorized_rider_sandbox_client.get_current_ride_details()
    assert response.status_code == codes.ok

    # assert response looks like ride details
    response = response.json
    assert EXPECTED_SHARED_RIDE_DETAILS_KEYS.issubset(response)
    assert response.get('status') == 'processing'


@uber_vcr.use_cassette()
def test_update_ride_destination(authorized_rider_sandbox_client):
    """Test to update the trip destination."""
    response = authorized_rider_sandbox_client.update_ride(
        RIDE_ID,
        end_latitude=UPDATE_LAT,
        end_longitude=UPDATE_LNG,
    )
    assert response.status_code == codes.no_content


@uber_vcr.use_cassette()
def test_update_ride_destination_with_places(authorized_rider_sandbox_client):
    """Test to update the trip destination with a place ID."""
    response = authorized_rider_sandbox_client.update_ride(
        RIDE_ID,
        end_place_id='work',
    )
    assert response.status_code == codes.no_content


@uber_vcr.use_cassette()
def test_get_ride_map(authorized_rider_sandbox_client):
    """Test to fetch ride map with access token."""
    response = authorized_rider_sandbox_client.get_ride_map(RIDE_ID)
    assert response.status_code == codes.ok

    # assert response looks like map
    response = response.json
    assert EXPECTED_RIDE_MAP_KEYS.issubset(response)


@uber_vcr.use_cassette()
def test_update_sandbox_ride(authorized_rider_sandbox_client):
    """Test to update sandbox ride status with access token."""
    response = authorized_rider_sandbox_client.update_sandbox_ride(
        ride_id=RIDE_ID,
        new_status='accepted',
    )
    assert response.status_code == codes.no_content


@uber_vcr.use_cassette()
def test_cancel_ride(authorized_rider_sandbox_client):
    """Test to cancel ride with access token."""
    response = authorized_rider_sandbox_client.cancel_ride(RIDE_ID)
    assert response.status_code == codes.no_content


@uber_vcr.use_cassette()
def test_cancel_current_ride(authorized_rider_sandbox_client):
    """Test to cancel the current ride with access token."""
    response = authorized_rider_sandbox_client.cancel_current_ride()
    assert response.status_code == codes.no_content


@uber_vcr.use_cassette()
def test_get_ride_receipt(authorized_rider_sandbox_client):
    """Test to fetch ride receipt with access token."""
    response = authorized_rider_sandbox_client.get_ride_receipt(RIDE_ID)
    assert response.status_code == codes.ok

    # assert response looks like ride receipt
    response = response.json
    assert EXPECTED_RECEIPT_KEYS.issubset(response)
    charges = response.get('charge_adjustments')

    for charge in charges:
        assert EXPECTED_INDIVIDUAL_CHARGE_KEYS.issubset(charge)


@uber_vcr.use_cassette()
def test_update_sandbox_product(authorized_rider_sandbox_client):
    """Test to update sandbox ride status with access token."""
    response = authorized_rider_sandbox_client.update_sandbox_product(
        product_id=NON_UFP_SURGE_PRODUCT_ID,
        surge_multiplier=2,
    )
    assert response.status_code == codes.no_content


@uber_vcr.use_cassette()
def test_request_ride_with_surge(authorized_rider_sandbox_client):
    """Test raising a SurgeError when requesting a ride with surge."""
    with raises(SurgeError):
        authorized_rider_sandbox_client.request_ride(
            product_id=NON_UFP_SURGE_PRODUCT_ID,
            start_latitude=NON_UFP_SURGE_START_LAT,
            start_longitude=NON_UFP_SURGE_START_LNG,
            end_latitude=NON_UFP_SURGE_END_LAT,
            end_longitude=NON_UFP_SURGE_END_LNG,
        )


def test_surge_error_formation(http_surge_error):
    """Test HTTP surge error response converted to SurgeError correctly."""
    surge_error = SurgeError(http_surge_error, 'msg')

    assert str(surge_error) == 'msg'
    assert isinstance(surge_error.errors, list)
    assert isinstance(surge_error.meta, dict)

    error_details = surge_error.errors[0]
    expected_code = 'surge'
    expected_description = http.ERROR_CODE_DESCRIPTION_DICT[expected_code]

    assert isinstance(error_details, ErrorDetails)
    assert error_details.status == http.STATUS_CONFLICT
    assert error_details.code == expected_code
    assert error_details.title == expected_description

    assert surge_error.surge_confirmation_id == SURGE_ID
    assert surge_error.surge_confirmation_href == SURGE_HREF.format(SURGE_ID)


@uber_vcr.use_cassette()
def test_set_home_address(authorized_rider_sandbox_client):
    """Test to update a user's home address with an access token."""
    response = authorized_rider_sandbox_client.set_home_address(HOME_ADDRESS)
    assert response.status_code == codes.ok

    # assert response looks like places details
    response = response.json
    assert EXPECTED_PLACE_KEYS.issubset(response)
    assert response.get('address') == FULL_HOME_ADDRESS


@uber_vcr.use_cassette()
def test_get_home_address(authorized_rider_sandbox_client):
    """Test to fetch a user's home address with an access token."""
    response = authorized_rider_sandbox_client.get_home_address()
    assert response.status_code == codes.ok

    # assert response looks like places details
    response = response.json
    assert EXPECTED_PLACE_KEYS.issubset(response)
    assert response.get('address') == FULL_HOME_ADDRESS


@uber_vcr.use_cassette()
def test_set_work_address(authorized_rider_sandbox_client):
    """Test to update a user's work address with an access token."""
    response = authorized_rider_sandbox_client.set_work_address(WORK_ADDRESS)
    assert response.status_code == codes.ok

    # assert response looks like places details
    response = response.json
    assert EXPECTED_PLACE_KEYS.issubset(response)
    assert response.get('address') == FULL_WORK_ADDRESS


@uber_vcr.use_cassette()
def test_get_work_address(authorized_rider_sandbox_client):
    """Test to fetch a user's work address with an access token."""
    response = authorized_rider_sandbox_client.get_work_address()
    assert response.status_code == codes.ok

    # assert response looks like places details
    response = response.json
    assert EXPECTED_PLACE_KEYS.issubset(response)
    assert response.get('address') == FULL_WORK_ADDRESS


@uber_vcr.use_cassette()
def test_get_payment_methods(authorized_rider_sandbox_client):
    """Test to get a list of payment methods with an access token."""
    response = authorized_rider_sandbox_client.get_payment_methods()
    assert response.status_code == codes.ok

    response = response.json
    payments = response.get('payment_methods')

    for payment in payments:
        assert EXPECTED_PAYMENT_KEYS.issubset(payment)


@uber_vcr.use_cassette()
def test_get_driver_profile(authorized_driver_production_client):
    """Test to fetch driver profile with access token."""
    response = authorized_driver_production_client.get_driver_profile()
    assert response.status_code == codes.ok

    # assert response looks like driver profile
    response = response.json
    assert EXPECTED_DRIVER_PROFILE_KEYS.issubset(response)


@uber_vcr.use_cassette()
def test_get_driver_trips(authorized_driver_production_client):
    """Test to fetch driver trips with access token."""
    response = authorized_driver_production_client.get_driver_trips()
    assert response.status_code == codes.ok

    response = response.json
    trips = response.get('trips')

    for trip in trips:
        assert EXPECTED_DRIVER_TRIPS_KEYS.issubset(trip)


@uber_vcr.use_cassette()
def test_get_driver_payments(authorized_driver_production_client):
    """Test to fetch driver payments with access token."""
    response = authorized_driver_production_client.get_driver_payments()
    assert response.status_code == codes.ok

    response = response.json
    payments = response.get('payments')

    for payment in payments:
        assert EXPECTED_DRIVER_PAYMENTS_KEYS.issubset(payment)



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""An internal module to handle OAuth 2.0 Authorization.

There are three ways you may obtain an access token:

    - Authorization Code Grant
    - Implicit Grant
    - Client Credentials Grant

Each OAuth 2.0 grant uses your app credentials to start an
authorization process with Uber. Upon successful authorization,
a Session is created, which stores the OAuth 2.0 credentials.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import OrderedDict
from random import SystemRandom
from requests import codes
from requests import post
from string import ascii_letters
from string import digits

try:
    from urllib.parse import parse_qs
    from urllib.parse import urlparse
except ImportError:
    from urlparse import parse_qs
    from urlparse import urlparse

from uber_rides.errors import ClientError
from uber_rides.errors import UberIllegalState
from uber_rides.session import OAuth2Credential
from uber_rides.session import Session
from uber_rides.utils import auth
from uber_rides.utils.request import build_url


class OAuth2(object):
    """The parent class for all OAuth 2.0 grant types."""

    def __init__(self, client_id, scopes):
        """Initialize OAuth 2.0 Class.

        Parameters
            client_id (str)
                Your app's Client ID.
            scopes (set)
                Set of permission scopes to request.
                (e.g. {'profile', 'history'}) Keep this list minimal so
                users feel safe granting your app access to their information.
        """
        self.client_id = client_id
        self.scopes = scopes

    def _build_authorization_request_url(
        self,
        response_type,
        redirect_url,
        state=None
    ):
        """Form URL to request an auth code or access token.

        Parameters
            response_type (str)
                Either 'code' (Authorization Code Grant) or
                'token' (Implicit Grant)
            redirect_url (str)
                The URL that the Uber server will redirect the user to after
                finishing authorization. The redirect must be HTTPS-based and
                match the URL you registered your application with. Localhost
                URLs are permitted and can be either HTTP or HTTPS.
            state (str)
                Optional CSRF State token to send to server.

        Returns
            (str)
                The fully constructed authorization request URL.

        Raises
            UberIllegalState (ApiError)
                Raised if response_type parameter is invalid.
        """
        if response_type not in auth.VALID_RESPONSE_TYPES:
            message = '{} is not a valid response type.'
            raise UberIllegalState(message.format(response_type))

        args = OrderedDict([
            ('scope', ' '.join(self.scopes)),
            ('state', state),
            ('redirect_uri', redirect_url),
            ('response_type', response_type),
            ('client_id', self.client_id),
        ])

        return build_url(auth.AUTH_HOST, auth.AUTHORIZE_PATH, args)

    def _extract_query(self, redirect_url):
        """Extract query parameters from a url.

        Parameters
            redirect_url (str)
                The full URL that the Uber server redirected to after
                the user authorized your app.

        Returns
            (dict)
                A dictionary of query parameters.
        """
        qs = urlparse(redirect_url)

        # Implicit Grant redirect_urls have data after fragment identifier (#)
        # All other redirect_urls return data after query identifier (?)
        qs = qs.fragment if isinstance(self, ImplicitGrant) else qs.query

        query_params = parse_qs(qs)
        query_params = {qp: query_params[qp][0] for qp in query_params}

        return query_params


class AuthorizationCodeGrant(OAuth2):
    """Class for Authorization Code Grant type.

    The authorization code grant type is used to obtain both access
    tokens and refresh tokens and is optimized for confidential clients.

    It involves a two-step authorization process. The first step is having
    the user authorize your app. The second involves getting an OAuth 2.0
    access token from Uber.
    """

    def __init__(
        self,
        client_id,
        scopes,
        client_secret,
        redirect_url,
        state_token=None,
    ):
        """Initialize AuthorizationCodeGrant Class.

        Parameters
            client_id (str)
                Your app's Client ID.
            scopes (set)
                Set of permission scopes to request.
                (e.g. {'profile', 'history'}) Keep this list minimal so
                users feel safe granting your app access to their information.
            client_secret (str)
                Your app's Client Secret.
            redirect_url (str)
                The URL that the Uber server will redirect the user to after
                finishing authorization. The redirect must be HTTPS-based and
                match the URL you registered your application with. Localhost
                URLs are permitted and can be either HTTP or HTTPS.
            state_token (str)
                The CSRF State Token used to create an authorization.
        """
        super(AuthorizationCodeGrant, self).__init__(client_id, scopes)
        self.redirect_url = redirect_url
        self.client_secret = client_secret

        # generate or set state token if not false
        self.state_token = False
        if state_token is None:
            self.state_token = self._generate_state_token()
        else:
            if state_token is not False:
                self.state_token = state_token

    def _generate_state_token(self, length=32):
        """Generate CSRF State Token.

        CSRF State Tokens are passed as a parameter in the authorization
        URL and are checked when receiving responses from the Uber Auth
        server to prevent request forgery.
        """
        choices = ascii_letters + digits
        return ''.join(SystemRandom().choice(choices) for _ in range(length))

    def get_authorization_url(self):
        """Start the Authorization Code Grant process.

        This function starts the OAuth 2.0 authorization process and builds an
        authorization URL. You should redirect your user to this URL, where
        they can grant your application access to their Uber account.

        Returns
            (str)
                The fully constructed authorization request URL.
                Tell the user to visit this URL and approve your app.
        """
        return self._build_authorization_request_url(
            response_type=auth.CODE_RESPONSE_TYPE,
            redirect_url=self.redirect_url,
            state=self.state_token,
        )

    def _verify_query(self, query_params):
        """Verify response from the Uber Auth server.

        Parameters
            query_params (dict)
                Dictionary of query parameters attached to your redirect URL
                after user approved your app and was redirected.

        Returns
            authorization_code (str)
                Code received when user grants your app access. Use this code
                to request an access token.

        Raises
            UberIllegalState (ApiError)
                Thrown if the redirect URL was missing parameters or if the
                given parameters were not valid.
        """
        error_message = None

        if self.state_token is not False:
            # Check CSRF State Token against state token from GET request
            received_state_token = query_params.get('state')
            if received_state_token is None:
                error_message = 'Bad Request. Missing state parameter.'
                raise UberIllegalState(error_message)

            if self.state_token != received_state_token:
                error_message = 'CSRF Error. Expected {}, got {}'
                error_message = error_message.format(
                    self.state_token,
                    received_state_token,
                )
                raise UberIllegalState(error_message)

        # Verify either 'code' or 'error' parameter exists
        error = query_params.get('error')
        authorization_code = query_params.get(auth.CODE_RESPONSE_TYPE)

        if error and authorization_code:
            error_message = (
                'Code and Error query params code and error '
                'can not both be set.'
            )
            raise UberIllegalState(error_message)

        if error is None and authorization_code is None:
            error_message = 'Neither query parameter code or error is set.'
            raise UberIllegalState(error_message)

        if error:
            raise UberIllegalState(error)

        return authorization_code

    def get_session(self, redirect_url):
        """Complete the Authorization Code Grant process.

        The redirect URL received after the user has authorized
        your application contains an authorization code. Use this
        authorization code to request an access token.

        Parameters
            redirect_url (str)
                The full URL that the Uber server redirected to after
                the user authorized your app.

        Returns
            (Session)
                A Session object with OAuth 2.0 credentials.
        """
        query_params = self._extract_query(redirect_url)
        authorization_code = self._verify_query(query_params)

        response = _request_access_token(
            grant_type=auth.AUTHORIZATION_CODE_GRANT,
            client_id=self.client_id,
            client_secret=self.client_secret,
            code=authorization_code,
            redirect_url=self.redirect_url,
        )

        oauth2credential = OAuth2Credential.make_from_response(
            response=response,
            grant_type=auth.AUTHORIZATION_CODE_GRANT,
            client_id=self.client_id,
            client_secret=self.client_secret,
            redirect_url=self.redirect_url,
        )

        return Session(oauth2credential=oauth2credential)


class ImplicitGrant(OAuth2):
    """Class for Implicit Grant type.

    The implicit grant type is used to obtain access tokens and is optimized
    for public clients under a particular redirect URI. It does not
    refresh access tokens.

    Unlike the authorization code grant type, in which the client makes
    separate requests for authorization and access token, the client
    receives the access token as the result of the authorization request.
    """

    def __init__(self, client_id, scopes, redirect_url):
        """Initialize ImplicitGrant Class.

        Parameters
            client_id (str)
                Your app's Client ID.
            scopes (set)
                Set of permission scopes to request.
                (e.g. {'profile', 'history'}) Keep this list minimal so
                users feel safe granting your app access to their information.
            redirect_url (str)
                The URL that the Uber server will redirect the user to after
                finishing authorization. The redirect must be HTTPS-based and
                match the URL you registered your application with. Localhost
                URLs are permitted and can be either HTTP or HTTPS.
        """
        super(ImplicitGrant, self).__init__(client_id, scopes)
        self.redirect_url = redirect_url

    def get_authorization_url(self):
        """Build URL for authorization request.

        Returns
            (str)
                The fully constructed authorization request URL.
        """
        return self._build_authorization_request_url(
            response_type=auth.TOKEN_RESPONSE_TYPE,
            redirect_url=self.redirect_url,
        )

    def get_session(self, redirect_url):
        """Create Session to store credentials.

        Parameters
            redirect_url (str)
                The full URL that the Uber server redirected to after
                the user authorized your app.

        Returns
            (Session)
                A Session object with OAuth 2.0 credentials.

        Raises
            UberIllegalState (APIError)
                Raised if redirect URL contains an error.
        """
        query_params = self._extract_query(redirect_url)

        error = query_params.get('error')
        if error:
            raise UberIllegalState(error)

        # convert space delimited string to set
        scopes = query_params.get('scope')
        scopes_set = {scope for scope in scopes.split()}

        oauth2credential = OAuth2Credential(
            client_id=self.client_id,
            redirect_url=self.redirect_url,
            access_token=query_params.get('access_token'),
            expires_in_seconds=query_params.get('expires_in'),
            scopes=scopes_set,
            grant_type=auth.IMPLICIT_GRANT,
        )

        return Session(oauth2credential=oauth2credential)


class ClientCredentialGrant(OAuth2):
    """Class for Client Credential Grant type.

    The client credential grant type is used to request an access token using
    only its client credentials.

    The client credentials grant type must only be used by confidential
    clients or when the client is requesting access to protected resources
    under its control.
    """

    def __init__(self, client_id, scopes, client_secret):
        """Initialize ClientCredential Class.

        Parameters
            client_id (str)
                Your app's Client ID.
            scopes (set)
                Set of permission scopes to request.
                (e.g. {'profile', 'history'}) Keep this list minimal so
                users feel safe granting your app access to their information.
            client_secret (str)
                Your app's Client Secret. This must be kept confidential.
        """
        super(ClientCredentialGrant, self).__init__(client_id, scopes)
        self.client_secret = client_secret

    def get_session(self):
        """Create Session to store credentials.

        Returns
            (Session)
                A Session object with OAuth 2.0 credentials.
        """
        response = _request_access_token(
            grant_type=auth.CLIENT_CREDENTIAL_GRANT,
            client_id=self.client_id,
            client_secret=self.client_secret,
            scopes=self.scopes,
        )

        oauth2credential = OAuth2Credential.make_from_response(
            response=response,
            grant_type=auth.CLIENT_CREDENTIAL_GRANT,
            client_id=self.client_id,
            client_secret=self.client_secret,
        )

        return Session(oauth2credential=oauth2credential)


def _request_access_token(
    grant_type,
    client_id=None,
    client_secret=None,
    scopes=None,
    code=None,
    redirect_url=None,
    refresh_token=None
):
    """Make an HTTP POST to request an access token.

    Parameters
        grant_type (str)
            Either 'client_credientials' (Client Credentials Grant)
            or 'authorization_code' (Authorization Code Grant).
        client_id (str)
            Your app's Client ID.
        client_secret (str)
            Your app's Client Secret.
        scopes (set)
            Set of permission scopes to request.
            (e.g. {'profile', 'history'})
        code (str)
            The authorization code to switch for an access token.
            Only used in Authorization Code Grant.
        redirect_url (str)
            The URL that the Uber server will redirect to.
        refresh_token (str)
            Refresh token used to get a new access token.
            Only used for Authorization Code Grant.

    Returns
        (requests.Response)
            Successful HTTP response from a 'POST' to request
            an access token.

    Raises
        ClientError (APIError)
            Thrown if there was an HTTP error.
    """
    url = build_url(auth.AUTH_HOST, auth.ACCESS_TOKEN_PATH)

    if isinstance(scopes, set):
        scopes = ' '.join(scopes)

    args = {
        'grant_type': grant_type,
        'client_id': client_id,
        'client_secret': client_secret,
        'scope': scopes,
        'code': code,
        'redirect_uri': redirect_url,
        'refresh_token': refresh_token,
    }

    response = post(url=url, data=args)

    if response.status_code == codes.ok:
        return response

    message = 'Failed to request access token: {}.'
    message = message.format(response.reason)
    raise ClientError(response, message)


def refresh_access_token(credential):
    """Use a refresh token to request a new access token.

    Not suported for access tokens obtained via Implicit Grant.

    Parameters
        credential (OAuth2Credential)
            An authorized user's OAuth 2.0 credentials.

    Returns
        (Session)
            A new Session object with refreshed OAuth 2.0 credentials.

    Raises
        UberIllegalState (APIError)
            Raised if OAuth 2.0 grant type does not support
            refresh tokens.
    """
    if credential.grant_type == auth.AUTHORIZATION_CODE_GRANT:
        response = _request_access_token(
            grant_type=auth.REFRESH_TOKEN,
            client_id=credential.client_id,
            client_secret=credential.client_secret,
            redirect_url=credential.redirect_url,
            refresh_token=credential.refresh_token,
        )

        oauth2credential = OAuth2Credential.make_from_response(
            response=response,
            grant_type=credential.grant_type,
            client_id=credential.client_id,
            client_secret=credential.client_secret,
            redirect_url=credential.redirect_url,
        )

        return Session(oauth2credential=oauth2credential)

    elif credential.grant_type == auth.CLIENT_CREDENTIAL_GRANT:
        response = _request_access_token(
            grant_type=auth.CLIENT_CREDENTIAL_GRANT,
            client_id=credential.client_id,
            client_secret=credential.client_secret,
            scopes=credential.scopes,
        )

        oauth2credential = OAuth2Credential.make_from_response(
            response=response,
            grant_type=credential.grant_type,
            client_id=credential.client_id,
            client_secret=credential.client_secret,
        )

        return Session(oauth2credential=oauth2credential)

    message = '{} Grant Type does not support Refresh Tokens.'
    message = message.format(credential.grant_type)
    raise UberIllegalState(message)


def revoke_access_token(credential):
    """Revoke an access token.

    All future requests with the access token will be invalid.

    Parameters
        credential (OAuth2Credential)
            An authorized user's OAuth 2.0 credentials.

    Raises
        ClientError (APIError)
            Thrown if there was an HTTP error.
    """
    url = build_url(auth.AUTH_HOST, auth.REVOKE_PATH)

    args = {
        'token': credential.access_token,
        'client_id': credential.client_id,
        'client_secret': credential.client_secret,
    }

    response = post(url=url, params=args)

    if response.status_code == codes.ok:
        return

    message = 'Failed to revoke access token: {}.'
    message = message.format(response.reason)
    raise ClientError(response, message)



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
from requests import Request

try:
    from urllib.parse import quote
    from urllib.parse import urlencode
    from urllib.parse import urljoin
except ImportError:
    from urllib import quote
    from urllib import urlencode
    from urlparse import urljoin

from uber_rides.utils.handlers import error_handler
from uber_rides.utils import http


def generate_data(method, args):
    """Assign arguments to body or URL of an HTTP request.

    Parameters
        method (str)
            HTTP Method. (e.g. 'POST')
        args (dict)
            Dictionary of data to attach to each Request.
            e.g. {'latitude': 37.561, 'longitude': -122.742}

    Returns
        (str or dict)
            Either params containing the dictionary of arguments
            or data containing arugments in JSON-formatted string.
    """
    data = {}
    params = {}

    if method in http.BODY_METHODS:
        data = dumps(args)
    else:
        params = args
    return data, params


def generate_prepared_request(method, url, headers, data, params, handlers):
    """Add handlers and prepare a Request.

    Parameters
        method (str)
            HTTP Method. (e.g. 'POST')
        headers (dict)
            Headers to send.
        data (JSON-formatted str)
            Body to attach to the request.
        params (dict)
            Dictionary of URL parameters to append to the URL.
        handlers (list)
            List of callback hooks, for error handling.

    Returns
        (requests.PreparedRequest)
            The fully mutable PreparedRequest object,
            containing the exact bytes to send to the server.
    """
    request = Request(
        method=method,
        url=url,
        headers=headers,
        data=data,
        params=params,
    )

    handlers.append(error_handler)

    for handler in handlers:
        request.register_hook('response', handler)

    return request.prepare()


def build_url(host, path, params=None):
    """Build a URL.

    This method encodes the parameters and adds them
    to the end of the base URL, then adds scheme and hostname.

    Parameters
        host (str)
            Base URL of the Uber Server that handles API calls.
        path (str)
            Target path to add to the host (e.g. 'v1.2/products').
        params (dict)
            Optional dictionary of parameters to add to the URL.

    Returns
        (str)
            The fully formed URL.
    """
    path = quote(path)
    params = params or {}

    if params:
        path = '/{}?{}'.format(path, urlencode(params))
    else:
        path = '/{}'.format(path)

    if not host.startswith(http.URL_SCHEME):
        host = '{}{}'.format(http.URL_SCHEME, host)

    return urljoin(host, path)



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

URL_SCHEME = 'https://'

ALLOWED_METHODS = frozenset(['GET', 'POST', 'PUT', 'DELETE', 'PATCH'])
BODY_METHODS = frozenset(['POST', 'PUT', 'PATCH'])
QUERY_METHODS = frozenset(['GET', 'DELETE'])

DEFAULT_CONTENT_HEADERS = {'content-type': 'application/json'}

STATUS_OK = 200
STATUS_UNAUTHORIZED = 401
STATUS_CONFLICT = 409
STATUS_UNPROCESSABLE_ENTITY = 422
STATUS_INTERNAL_SERVER_ERROR = 500
STATUS_SERVICE_UNAVAILABLE = 503

ERROR_CODE_DESCRIPTION_DICT = {
    'distance_exceeded': 'Distance between two points exceeds 100 miles.',
    'unauthorized': 'Invalid OAuth 2.0 credentials provided.',
    'validation_failed': 'Invalid request.',
    'internal_server_error': 'Unexpected internal server error occurred.',
    'service_unavailable': 'Service temporarily unavailable.',
    'surge': 'Surge pricing is in effect.',
    'same_pickup_dropoff': 'Pickup and Dropoff can\'t be the same.',
}

VALID_TOKEN_TYPES = frozenset(['Token', 'Bearer'])



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Python client for the Uber API.

This client is designed to make calls to the Uber API.
An UberRidesClient is instantiated with a Session which holds either
your server token or OAuth 2.0 credentials. Your usage of this
module might look like:

    client = UberRidesClient(session)
    products = client.get_products(latitude, longitude)
    profile = client.get_user_profile()
    ride_details = client.get_ride_details(ride_id)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import OrderedDict
from requests import codes

from uber_rides.auth import refresh_access_token
from uber_rides.auth import revoke_access_token
from uber_rides.errors import ClientError
from uber_rides.errors import UberIllegalState
from uber_rides.request import Request
from uber_rides.utils import auth


VALID_PRODUCT_STATUS = frozenset([
    'processing',
    'accepted',
    'arriving',
    'in_progress',
    'driver_canceled',
    'completed',
])

PRODUCTION_HOST = 'api.uber.com'
SANDBOX_HOST = 'sandbox-api.uber.com'


class UberRidesClient(object):
    """Class to make calls to the Uber API."""

    def __init__(self, session, sandbox_mode=False):
        """Initialize an UberRidesClient.

        Parameters
            session (Session)
                The Session object containing access credentials.
            sandbox_mode (bool)
                Default (False) is not using sandbox mode.
        """
        self.session = session
        self.api_host = SANDBOX_HOST if sandbox_mode else PRODUCTION_HOST

    def _api_call(self, method, target, args=None):
        """Create a Request object and execute the call to the API Server.

        Parameters
            method (str)
                HTTP request (e.g. 'POST').
            target (str)
                The target URL with leading slash (e.g. '/v1.2/products').
            args (dict)
                Optional dictionary of arguments to attach to the request.

        Returns
            (Response)
                The server's response to an HTTP request.
        """
        self.refresh_oauth_credential()
        handlers = [surge_handler]
        request = Request(
            auth_session=self.session,
            api_host=self.api_host,
            method=method,
            path=target,
            handlers=handlers,
            args=args,
        )

        return request.execute()

    def get_products(self, latitude, longitude):
        """Get information about the Uber products offered at a given location.

        Parameters
            latitude (float)
                The latitude component of a location.
            longitude (float)
                The longitude component of a location.

        Returns
            (Response)
                A Response object containing available products information.
        """
        args = OrderedDict([
            ('latitude', latitude),
            ('longitude', longitude),
        ])

        return self._api_call('GET', 'v1.2/products', args=args)

    def get_product(self, product_id):
        """Get information about a specific Uber product.

        Parameters
            product_id (str)
                Unique identifier representing a specific product for a
                given location.

        Returns
            (Response)
                A Response containing information about a specific product.
        """
        endpoint = 'v1.2/products/{}'.format(product_id)
        return self._api_call('GET', endpoint)

    def get_price_estimates(
        self,
        start_latitude,
        start_longitude,
        end_latitude,
        end_longitude,
        seat_count=None,
    ):
        """Get price estimates for products at a given location.

        Parameters
            start_latitude (float)
                The latitude component of a start location.
            start_longitude (float)
                The longitude component of a start location.
            end_latitude (float)
                The latitude component of a end location.
            end_longitude (float)
                The longitude component of a end location.
            seat_count (int)
                The number of seats required for uberPOOL.
                Default and maximum value is 2.

        Returns
            (Response)
                A Response object containing each product's price estimates.
        """
        args = OrderedDict([
            ('start_latitude', start_latitude),
            ('start_longitude', start_longitude),
            ('end_latitude', end_latitude),
            ('end_longitude', end_longitude),
            ('seat_count', seat_count),
        ])

        return self._api_call('GET', 'v1.2/estimates/price', args=args)

    def get_pickup_time_estimates(
        self,
        start_latitude,
        start_longitude,
        product_id=None,
    ):
        """Get pickup time estimates for products at a given location.

        Parameters
            start_latitude (float)
                The latitude component of a start location.
            start_longitude (float)
                The longitude component of a start location.
            product_id (str)
                The unique ID of the product being requested. If none is
                provided, it will default to the cheapest product for the
                given location.

        Returns
            (Response)
                A Response containing each product's pickup time estimates.
        """
        args = OrderedDict([
            ('start_latitude', start_latitude),
            ('start_longitude', start_longitude),
            ('product_id', product_id),
        ])

        return self._api_call('GET', 'v1.2/estimates/time', args=args)

    def get_promotions(
        self,
        start_latitude,
        start_longitude,
        end_latitude,
        end_longitude,
    ):
        """Get information about the promotions available to a user.

        Parameters
            start_latitude (float)
                The latitude component of a start location.
            start_longitude (float)
                The longitude component of a start location.
            end_latitude (float)
                The latitude component of a end location.
            end_longitude (float)
                The longitude component of a end location.

        Returns
            (Response)
                A Response object containing available promotions.
        """

        args = OrderedDict([
            ('start_latitude', start_latitude),
            ('start_longitude', start_longitude),
            ('end_latitude', end_latitude),
            ('end_longitude', end_longitude)
        ])

        return self._api_call('GET', 'v1.2/promotions', args=args)

    def get_user_activity(self, offset=None, limit=None):
        """Get activity about the user's lifetime activity with Uber.

        Parameters
            offset (int)
                The integer offset for activity results. Default is 0.
            limit (int)
                Integer amount of results to return. Maximum is 50.
                Default is 5.

        Returns
            (Response)
                A Response object containing ride history.
        """
        args = {
            'offset': offset,
            'limit': limit,
        }

        return self._api_call('GET', 'v1.2/history', args=args)

    def get_user_profile(self):
        """Get information about the authorized Uber user.

        Returns
            (Response)
                A Response object containing account information.
        """
        return self._api_call('GET', 'v1.2/me')

    def apply_promotion_code(
        self,
        promotion_code=None,
    ):
        """Apply a promotion code to an Uber user.

        Parameters
            promotion_code (str)
                The unique promotion code to apply.

        Returns
            (Response)
                A Response object containing the applied promotion codes.
        """
        args = {
            'applied_promotions_codes': promotion_code
        }

        return self._api_call('PATCH', 'v1.2/me', args=args)

    def estimate_ride(
        self,
        product_id=None,
        start_latitude=None,
        start_longitude=None,
        start_place_id=None,
        end_latitude=None,
        end_longitude=None,
        end_place_id=None,
        seat_count=None,
    ):
        """Estimate ride details given a product, start, and end location.

        Only pickup time estimates and surge pricing information are provided
        if no end location is provided.

        Parameters
            product_id (str)
                The unique ID of the product being requested. If none is
                provided, it will default to the cheapest product for the
                given location.
            start_latitude (float)
                The latitude component of a start location.
            start_longitude (float)
                The longitude component of a start location.
            start_place_id (str)
                The beginning or pickup place ID. Only "home" or "work"
                is acceptable.
            end_latitude (float)
                Optional latitude component of a end location.
            end_longitude (float)
                Optional longitude component of a end location.
            end_place_id (str)
                The final or destination place ID. Only "home" or "work"
                is acceptable.
            seat_count (str)
                Optional Seat count for shared products. Default is 2.


        Returns
            (Response)
                A Response object containing fare id, time, price, and distance
                estimates for a ride.
        """
        args = {
            'product_id': product_id,
            'start_latitude': start_latitude,
            'start_longitude': start_longitude,
            'start_place_id': start_place_id,
            'end_latitude': end_latitude,
            'end_longitude': end_longitude,
            'end_place_id': end_place_id,
            'seat_count': seat_count
        }

        return self._api_call('POST', 'v1.2/requests/estimate', args=args)

    def request_ride(
        self,
        product_id=None,
        start_latitude=None,
        start_longitude=None,
        start_place_id=None,
        start_address=None,
        start_nickname=None,
        end_latitude=None,
        end_longitude=None,
        end_place_id=None,
        end_address=None,
        end_nickname=None,
        seat_count=None,
        fare_id=None,
        surge_confirmation_id=None,
        payment_method_id=None,
    ):
        """Request a ride on behalf of an Uber user.

        When specifying pickup and dropoff locations, you can either use
        latitude/longitude pairs or place ID (but not both).

        Parameters
            product_id (str)
                The unique ID of the product being requested. If none is
                provided, it will default to the cheapest product for the
                given location.
            start_latitude (float)
                Optional latitude component of a start location.
            start_longitude (float)
                Optional longitude component of a start location.
            start_place_id (str)
                The beginning or pickup place ID. Only "home" or "work"
                is acceptable.
            start_address (str)
                Optional pickup address.
            start_nickname (str)
                Optional pickup nickname label.
            end_latitude (float)
                Optional latitude component of a end location.
            end_longitude (float)
                Optional longitude component of a end location.
            end_place_id (str)
                The final or destination place ID. Only "home" or "work"
                is acceptable.
            end_address (str)
                Optional destination address.
            end_nickname (str)
                Optional destination nickname label.
            seat_count (int)
                Optional seat count for shared products.
            fare_id (str)
                Optional fare_id for shared products.
            surge_confirmation_id (str)
                Optional unique identifier of the surge session for a user.
            payment_method_id (str)
                Optional unique identifier of the payment method selected
                by a user. If set, the trip will be requested using this
                payment method. If not, the trip will be requested with the
                user's last used payment method.

        Returns
            (Response)
                A Response object containing the ride request ID and other
                details about the requested ride.

        Raises
            SurgeError (ClientError)
                Thrown when the requested product is currently surging.
                User must confirm surge price through surge_confirmation_href.
        """
        args = {
            'product_id': product_id,
            'start_latitude': start_latitude,
            'start_longitude': start_longitude,
            'start_place_id': start_place_id,
            'start_address': start_address,
            'start_nickname': start_nickname,
            'end_latitude': end_latitude,
            'end_longitude': end_longitude,
            'end_place_id': end_place_id,
            'end_address': end_address,
            'end_nickname': end_nickname,
            'surge_confirmation_id': surge_confirmation_id,
            'payment_method_id': payment_method_id,
            'seat_count': seat_count,
            'fare_id': fare_id
        }

        return self._api_call('POST', 'v1.2/requests', args=args)

    def get_ride_details(self, ride_id):
        """Get status details about an ongoing or past ride.

        Params
            ride_id (str)
                The unique ID of the Ride Request.

        Returns
            (Response)
                A Response object containing the ride's
                status, location, driver, and other details.
        """
        endpoint = 'v1.2/requests/{}'.format(ride_id)
        return self._api_call('GET', endpoint)

    def get_current_ride_details(self):
        """Get status details for an ongoing ride.

        This method behaves like get_ride_details by default (only returns
        details about trips your app requested). If your app has the
        `all_trips` scope, however, trip details will be returned for trips
        irrespective of which application initiated them.

        Returns
            (Response)
                A Response object containing details about a user's
                current trip - if any.
        """
        return self._api_call('GET', 'v1.2/requests/current')

    def update_ride(
        self,
        ride_id,
        end_latitude=None,
        end_longitude=None,
        end_place_id=None,
    ):
        """Update an ongoing ride's destination.

        To specify a new dropoff location, you can either use a
        latitude/longitude pair or place ID (but not both).

        Params
            ride_id (str)
                The unique ID of the Ride Request.
            end_latitude (float)
                The latitude component of a end location.
            end_longitude (float)
                The longitude component of a end location.
            end_place_id (str)
                The final or destination place ID. This is the name of an
                Uber saved place. Only "home" or "work" is acceptable.
            end_address (str)
                The final or destination address.
            end_nickname (str)
                The final or destination nickname label.

        Returns
            (Response)
                The Response with successful status_code
                if the ride's destination was updated.
        """
        args = {}
        if end_latitude is not None:
            args.update({'end_latitude': end_latitude})
        if end_longitude is not None:
            args.update({'end_longitude': end_longitude})
        if end_place_id is not None:
            args.update({'end_place_id': end_place_id})

        endpoint = 'v1.2/requests/{}'.format(ride_id)
        return self._api_call('PATCH', endpoint, args=args)

    def cancel_ride(self, ride_id):
        """Cancel an ongoing ride on behalf of a user.

        Params
            ride_id (str)
                The unique ID of the Ride Request.

        Returns
            (Response)
                A Response object with successful status_code
                if ride was canceled.
        """
        endpoint = 'v1.2/requests/{}'.format(ride_id)
        return self._api_call('DELETE', endpoint)

    def cancel_current_ride(self):
        """Cancel the user's current trip.

        This method behaves like cancel_ride, except you don't need
        to specify a request_id.

        Returns
            (Response)
                A Response object with successful status_code
                if ride was canceled.
        """
        return self._api_call('DELETE', 'v1.2/requests/current')

    def get_ride_map(self, ride_id):
        """Get a map with a visual representation of a Request.

        Maps are only available after a ride has been accepted by a
        driver and is in the `accepted` state. Attempting to get a map
        before that will result in a 404 error.

        Params
            ride_id (str)
                The unique ID of the Ride Request.

        Returns
            (Response)
                A Response object with a link to a map.
        """
        endpoint = 'v1.2/requests/{}/map'.format(ride_id)
        return self._api_call('GET', endpoint)

    def get_ride_receipt(self, ride_id):
        """Get receipt information from a completed ride.

        Params
            ride_id (str)
                The unique ID of the Ride Request.

        Returns
            (Response)
                A Response object containing the charges for
                the given ride.
        """
        endpoint = 'v1.2/requests/{}/receipt'.format(ride_id)
        return self._api_call('GET', endpoint)

    def update_sandbox_ride(self, ride_id, new_status):
        """Update the status of an ongoing sandbox request.

        Params
            ride_id (str)
                The unique ID of the Ride Request.
            new_status (str)
                Status from VALID_PRODUCT_STATUS.

        Returns
            (Response)
                A Response object with successful status_code
                if ride status was updated.
        """
        if new_status not in VALID_PRODUCT_STATUS:
            message = '{} is not a valid product status.'
            raise UberIllegalState(message.format(new_status))

        args = {'status': new_status}
        endpoint = 'v1.2/sandbox/requests/{}'.format(ride_id)
        return self._api_call('PUT', endpoint, args=args)

    def update_sandbox_product(
        self,
        product_id,
        surge_multiplier=None,
        drivers_available=None,
    ):
        """Update sandbox product availability.

        Params
            product_id (str)
                Unique identifier representing a specific product for a
                given location.
            surge_multiplier (float)
                Optional surge multiplier to manipulate pricing of product.
            drivers_available (bool)
                Optional boolean to manipulate availability of product.

        Returns
            (Response)
                The Response with successful status_code
                if product status was updated.
        """
        args = {
            'surge_multiplier': surge_multiplier,
            'drivers_available': drivers_available,
        }

        endpoint = 'v1.2/sandbox/products/{}'.format(product_id)
        return self._api_call('PUT', endpoint, args=args)

    def get_home_address(self):
        """Retrieve the saved home address for an Uber user.

        Returns
            (Response)
                A Response object with the home address - if one is set.
        """
        return self._api_call('GET', 'v1.2/places/home')

    def get_work_address(self):
        """Retrieve the saved work address for an Uber user.

        Returns
            (Response)
                A Response object with the work address - if one is set.
        """
        return self._api_call('GET', 'v1.2/places/work')

    def set_home_address(self, address):
        """Update saved home address for an Uber user.

        Params
            address (str)
                The address that should be assigned to "home".

        Returns
            (Response)
                A Response object with the updated home address.
        """
        args = {'address': address}

        return self._api_call('PUT', 'v1.2/places/home', args=args)

    def set_work_address(self, address):
        """Update saved work address for an Uber user.

        Params
            address (str)
                The address that should be assigned to "work".

        Returns
            (Response)
                A Response object with the updated work address.
        """
        args = {'address': address}

        return self._api_call('PUT', 'v1.2/places/work', args=args)

    def get_payment_methods(self):
        """Retrieve a list of the user's available payment methods.

        Returns
            (Response)
                A Response object containing information about a user's
                payment methods.
        """
        return self._api_call('GET', 'v1.2/payment-methods')

    def refresh_oauth_credential(self):
        """Refresh session's OAuth 2.0 credentials if they are stale."""
        if self.session.token_type == auth.SERVER_TOKEN_TYPE:
            return

        credential = self.session.oauth2credential
        if credential.is_stale():
            refresh_session = refresh_access_token(credential)
            self.session = refresh_session

    def revoke_oauth_credential(self):
        """Revoke the session's OAuth 2.0 credentials."""
        if self.session.token_type == auth.SERVER_TOKEN_TYPE:
            return

        credential = self.session.oauth2credential
        revoke_access_token(credential)

    def get_rider_profile(self):
        """Get information about the authorized Uber user.

        Returns
            (Response)
                A Response object containing account information.
        """
        return self.get_user_profile()

    def get_driver_profile(self):
        """Get profile about the authorized Uber driver.

        Returns
            (Response)
                A Response object containing account information.
        """
        return self._api_call('GET', 'v1/partners/me')

    def get_driver_trips(self,
                         offset=None,
                         limit=None,
                         from_time=None,
                         to_time=None
                         ):
        """Get trips about the authorized Uber driver.

        Parameters
            offset (int)
                The integer offset for activity results. Offset the list of
                returned results by this amount. Default is zero.
            limit (int)
                Integer amount of results to return. Number of items to
                retrieve per page. Default is 10, maximum is 50.
            from_time (int)
                Unix timestamp of the start time to query. Queries starting
                from the first trip if omitted.
            to_time (int)
                Unix timestamp of the end time to query. Queries starting
                from the last trip if omitted.

        Returns
            (Response)
                A Response object containing trip information.
        """
        args = {
            'offset': offset,
            'limit': limit,
            'from_time': from_time,
            'to_time': to_time,
        }
        return self._api_call('GET', 'v1/partners/trips', args=args)

    def get_driver_payments(self,
                            offset=None,
                            limit=None,
                            from_time=None,
                            to_time=None
                            ):
        """Get payments about the authorized Uber driver.

        Parameters
            offset (int)
                The integer offset for activity results. Offset the list of
                returned results by this amount. Default is zero.
            limit (int)
                Integer amount of results to return. Number of items to
                retrieve per page. Default is 10, maximum is 50.
            from_time (int)
                Unix timestamp of the start time to query. Queries starting
                from the first trip if omitted.
            to_time (int)
                Unix timestamp of the end time to query. Queries starting
                from the last trip if omitted.

        Returns
            (Response)
                A Response object containing trip information.
        """
        args = {
            'offset': offset,
            'limit': limit,
            'from_time': from_time,
            'to_time': to_time,
        }
        return self._api_call('GET', 'v1/partners/payments', args=args)


def surge_handler(response, **kwargs):
    """Error Handler to surface 409 Surge Conflict errors.

    Attached as a callback hook on the Request object.

    Parameters
        response (requests.Response)
            The HTTP response from an API request.
        **kwargs
            Arbitrary keyword arguments.
    """
    if response.status_code == codes.conflict:
        json = response.json()
        errors = json.get('errors', [])
        error = errors[0] if errors else json.get('error')

        if error and error.get('code') == 'surge':
            raise SurgeError(response)

    return response


class SurgeError(ClientError):
    """Raise for 409 Surge Conflicts."""

    def __init__(self, response, message=None):
        """
        Parameters
            response (requests.Response)
                The HTTP response from an API request.
            message (str)
                An error message string. If one is not provided, the
                default message is used.
        """
        if not message:
            message = (
                'Surge pricing is currently in effect for this product. '
                'User must confirm surge by visiting the confirmation url.'
            )

        super(SurgeError, self).__init__(
            response=response,
            message=message,
        )

        surge_href, surge_id = self.adapt_meta(self.meta)
        self.surge_confirmation_href = surge_href
        self.surge_confirmation_id = surge_id

    def adapt_meta(self, meta):
        """Convert meta from error response to href and surge_id attributes."""

        surge = meta.get('surge_confirmation')
        href = surge.get('href')
        surge_id = surge.get('surge_confirmation_id')

        return href, surge_id



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from uber_rides.errors import ClientError
from uber_rides.errors import ServerError


def error_handler(response, **kwargs):
    """Error Handler to surface 4XX and 5XX errors.

    Attached as a callback hook on the Request object.

    Parameters
        response (requests.Response)
            The HTTP response from an API request.
        **kwargs
            Arbitrary keyword arguments.

    Raises
        ClientError (ApiError)
            Raised if response contains a 4XX status code.
        ServerError (ApiError)
            Raised if response contains a 5XX status code.

    Returns
        response (requests.Response)
            The original HTTP response from the API request.
    """
    try:
        body = response.json()
    except ValueError:
        body = {}
    status_code = response.status_code
    message = body.get('message', '')
    fields = body.get('fields', '')
    error_message = str(status_code) + ': ' + message + ' ' + str(fields)
    if 400 <= status_code <= 499:
        raise ClientError(response, error_message)

    elif 500 <= status_code <= 599:
        raise ServerError(response, error_message)

    return response



#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from setuptools import find_packages
from setuptools import setup

with open('README.rst') as f:
    readme = f.read()

setup(
    name='uber_rides',
    version='0.5.0',
    packages=find_packages(),
    description='Official Uber Rides API Python SDK',
    long_description=readme,
    url='https://github.com/uber/rides-python-sdk',
    license='MIT',
    author='Uber Technologies, Inc.',
    author_email='christinek@uber.com',
    install_requires=['requests', 'pyyaml'],
    extras_require={
        ':python_version == "2.7"': ['future'],
    },
    tests_require=['pytest', 'mock', 'vcrpy'],
    keywords=['uber', 'api', 'sdk', 'rides', 'library'],
)
